/*
 * PythonPlugin.cc
 *
 *  Created on: Jan 13, 2012
 *      Author: lacasta
 */
#if HAVE_CONFIG_H
#ifdef _WIN32
#include "config_win.h"
#else
#include <config.h>
#endif
#endif

#if defined(HAS_PYTHON)
#include <Python.h>
#include "python-compat.h"
#include "PythonPlugin.h"
#include "PythonLoader.h"
#include <glib.h>
#include <gmodule.h>

#include <vector>
#include <string>




extern "C"
{
    int find_functions_in_module(const char *module_name, std::vector<std::string> &vout);


    const gchar *g_module_check_init(GModule *mod)
{
    int argc = 0;
    char *argv[] = {0, 0};
    printf("### Loading Python module...\n");
    PythonLoader::getInstance().initialize(argc, argv);
    return NULL;
}

void g_module_unload(GModule *m)
{
    printf("### UnLoading Python module...\n");
}

void unload_module(PyObject *module)
{
    if (module)
        Py_DecRef(module);
}

void add_path(const gchar *path)
{
    PythonLoader::getInstance().add_path(path);
}

Plugin *load_plugin(const char *module, const char *obj)
{
    Plugin *P = new PythonPlugin(module, obj);
    return P;
}

void show_path()
{
    PythonLoader::getInstance().show_path();
}


int find_symbols_in_module(gchar *path, std::vector< std::string > &symb)
{
    gchar *fnam = g_path_get_basename(path);
    gchar *dnam = g_path_get_dirname(path);
    add_python_path(dnam);
    find_functions_in_module(fnam, symb);
    g_free(fnam);
    g_free(dnam);

    return 0;
}


}

std::string string_from_string(PyObject * ret)
{
    std::string rc;
    if ( ret && PyString_Check(ret))
    {
        int len = PyString_GET_SIZE(ret);
        char *data = PyString_AsString(ret);
        rc = std::string(data, len);
    }
    return rc;
}



PythonPlugin::PythonPlugin(const char *module, const char *hook) :
    _instance_(0),
    _newfile_(0), _sor_(0), _eor_(0), _newevt_(0), _newpoint_(0), _filter_(0), _data_format_(0)
{
    PyObject *func=0;
    PyObject *__module=0;
    if ( module && hook )
    {
        func = PythonLoader::getInstance().execute_function_from_module(module, hook, true, &__module);
        if (func)
        {
            _instance_ = func;
            Py_INCREF(func);
        }
    }
    _module_ = __module;
    if (_instance_)
    {
        _newfile_ = PyObject_GetAttrString(_instance_,"new_file");
        _sor_ = PyObject_GetAttrString(_instance_,"start_of_run");
        _eor_ = PyObject_GetAttrString(_instance_,"end_of_run");
        _newevt_ = PyObject_GetAttrString(_instance_,"new_event");
        _newpoint_ = PyObject_GetAttrString(_instance_,"new_point");
        _filter_ = PyObject_GetAttrString(_instance_,"filter_event");
        _data_format_ = PyObject_GetAttrString(_instance_,"get_data_format");
    }
}
PythonPlugin::~PythonPlugin()
{
    Py_XDECREF(_newfile_);
    Py_XDECREF(_sor_);
    Py_XDECREF(_eor_);
    Py_XDECREF(_newevt_);
    Py_XDECREF(_newpoint_);
    Py_XDECREF(_filter_);
    Py_XDECREF(_instance_);
}

void PythonPlugin::get_data_format(std::string &data_format)
{
    if ( ! _instance_  || ! _newfile_)
    {
        data_format.clear();
        return;
    }

    PyGILState_STATE gilstate;
    gilstate = PyGILState_Ensure();
    PyObject *ret = PyObject_CallFunctionObjArgs(_data_format_, 0);
    if (PyErr_Occurred())
    {
        print_python_error();
        PyErr_Print();
    }
    else
    {
        data_format = string_from_string(ret);
    }

    if (ret)
    {
        Py_DECREF(ret);
    }
    PyGILState_Release(gilstate);

    return;
}

void PythonPlugin::new_file(std::string &rc)
{
    if ( ! _instance_  || ! _newfile_)
        return;

    PyGILState_STATE gilstate;
    gilstate = PyGILState_Ensure();
    //PyObject *ret = PyObject_CallMethod(_instance_, "new_file", 0);
    PyObject *ret = PyObject_CallFunctionObjArgs(_newfile_, 0);
    if (PyErr_Occurred())
    {
        print_python_error();
        PyErr_Print();
    }
    else
    {
        rc = string_from_string(ret);
    }

    if (ret)
    {
        Py_DECREF(ret);
    }
    PyGILState_Release(gilstate);

    return;
}

int PythonPlugin::start_of_run(int run_type, int nevts, int sample_size)
{
    int rc = nevts;

    if ( ! _instance_ || ! _sor_)
        return rc;

    PyGILState_STATE gilstate;
    gilstate = PyGILState_Ensure();
    //PyObject *ret = PyObject_CallMethod(_instance_, "start_of_run", "ii", run_type, nevts;
    PyObject *ret = PyObject_CallFunction(_sor_,(char *)"ii", run_type, nevts);
    if (PyErr_Occurred())
    {
        print_python_error();
        PyErr_Print();
    }
    else
    {
        if ( ret && PyInt_Check(ret))
        {
            rc = PyInt_AsLong(ret);
        }
    }

    if (ret)
    {
        Py_DECREF(ret);
    }
    PyGILState_Release(gilstate);

    return rc;
}

void PythonPlugin::end_of_run(std::string &S)
{
    if ( ! _instance_  || !_eor_)
        return;

    PyGILState_STATE gilstate;
    gilstate = PyGILState_Ensure();
    //PyObject *ret = PyObject_CallMethod(_instance_, "end_of_run", 0);
    PyObject *ret = PyObject_CallFunctionObjArgs(_eor_,0);
    if (PyErr_Occurred())
    {
        print_python_error();
        PyErr_Print();
    }
    else
    {
        S = string_from_string(ret);
    }

    if (ret)
    {
        Py_DECREF(ret);
    }
    PyGILState_Release(gilstate);
}


bool PythonPlugin::new_event(int ievt)
{
    bool rc=false;

    if ( ! _instance_ || !_newevt_)
        return rc;

    PyGILState_STATE gilstate;
    gilstate = PyGILState_Ensure();
    //PyObject *ret = PyObject_CallMethod(_instance_, "new_event", "i", ievt);
    PyObject *ret = PyObject_CallFunction(_newevt_,(char *)"i", ievt);
    if (PyErr_Occurred())
    {
        print_python_error();
        PyErr_Print();
    }
    else
    {
        if ( ret && PyInt_Check(ret))
        {
            rc = PyInt_AsLong(ret)!=0;
        }
    }

    if (ret)
    {
        Py_DECREF(ret);
    }
    PyGILState_Release(gilstate);

    return rc;

}
void PythonPlugin::new_point(std::string &rc)
{
    if ( ! _instance_ || !_newpoint_)
        return;

    PyGILState_STATE gilstate;
    gilstate = PyGILState_Ensure();
    //PyObject *ret = PyObject_CallMethod(_instance_, "new_point", 0);
    PyObject *ret = PyObject_CallFunctionObjArgs(_newpoint_,0);
    if (PyErr_Occurred())
    {
        print_python_error();
        PyErr_Print();
    }
    else
    {
        rc = string_from_string(ret);
    }

    if (ret)
    {
        Py_DECREF(ret);
    }
    PyGILState_Release(gilstate);

    return;
}

int PythonPlugin::filter_event(const AlibavaData *evt, std::string &rc)
{
    if (!_instance_ || !_filter_)
        return 0;

    PyGILState_STATE gilstate;
    gilstate = PyGILState_Ensure();

    int ichip, nchip = evt->nchip();
    PyObject *buffer;
    PyObject *_time, *_temp, *_val;
    buffer = PyTuple_New(nchip);
    for (ichip = 0; ichip<nchip; ichip++)
    {
        /*
         * We send
         *   ( chip1, chip2 )
         *   where chip is
         *   chip = ( header, data)
         *   header: array of 16 header values
         *   data: array with 128 channel values
         */
        PyObject *chip = PyTuple_New(2);
        PyTuple_SetItem(buffer, ichip, chip);
        PyObject *header = PyTuple_New(16);
        PyObject *data = PyTuple_New(128);
        PyTuple_SetItem(chip, 0, header);
        PyTuple_SetItem(chip, 1, data);
        for (int item=0; item<16; item++)
        {
            PyTuple_SetItem(header, item, PyInt_FromLong((long) evt->chip(ichip)->header[item]));
        }
        for (int item=0; item<128; item++)
        {
            PyTuple_SetItem(data, item, PyInt_FromLong((long) evt->chip(ichip)->signal[item]));
        }
    }
    _time = PyInt_FromLong((long)evt->tdc());
    _temp = PyInt_FromLong((long)evt->temp());
    _val  = PyFloat_FromDouble((double)evt->value());

    //PyObject *ret = PyObject_CallMethod(_instance_,
    //                                    "filter_event",
    //                                    "ihdO", evt.time, evt.temp, evt.value, buffer);

    PyObject *ret = PyObject_CallFunctionObjArgs(_filter_,_time, _temp, _val, buffer, 0);
    if (PyErr_Occurred())
    {
        print_python_error();
        PyErr_Print();
    }
    else
    {
        rc = string_from_string(ret);
    }

    Py_DECREF(_time);
    Py_DECREF(_temp);
    Py_DECREF(_val);
    Py_DECREF(buffer);
    if (ret)
    {
        Py_DECREF(ret);
    }
    PyGILState_Release(gilstate);

    return (int)rc.size();
}

#endif
