/*
 * alibavaClient.cc
 *
 *  Created on: Jan 21, 2014
 *      Author: lacasta
 */
#include <iostream>
#include <iomanip>
#include <sstream>
#include <algorithm>

#include "alibavaClient.h"

std::ostream &operator<<(std::ostream &os, const ns1__Status &st)
{
    double rate = st.rate;
    std::string rate_units = " Hz";
    std::string thr_units;

    if (rate > 1.e6)
    {
        rate /= 1.e6;
        rate_units = "MHz";
    }
    else if (rate > 1000.)
    {
        rate /= 1000.;
        rate_units = "kHz";
    }

    os << "Evts " << std::setiosflags(std::ios::fixed | std::ios::right)
       << std::setw(7) << st.ntrigger << " time " << std::setw(6)
       << std::setprecision(1) << st.time << " rate " << std::setw(8)
       << std::setprecision(1) << rate << rate_units << " eff " << std::setw(5)
       << "[" << st.state << "]"
       << std::flush;
    return os;
}

Alibava::Alibava(const char *uri)
: server(0)
{
    if (uri)
        connect(uri);
}

void Alibava::connect(const char *uri)
{
    if (server)
        delete server;

    server = new alibavaSOAPProxy(uri);
}
Alibava::~Alibava()
{
    if (server)
        delete server;

    reset_histograms();
}

struct HstEraser : public std::unary_function<Histogram *, void>
{
        HstEraser() {}
        void operator() (Histogram *hst) {
            delete hst;
        }
};
void Alibava::reset_histograms()
{
    std::for_each(histograms.begin(), histograms.end(), HstEraser() );
    histograms.clear();
}

int Alibava::setDataFile(const char *ofile)
{
    if (!server)
        return 1;

    ns1__setDataFileResponse response;
    server->setDataFile(ofile, response);
    return response.setDataFileResponse;
}

int Alibava::setParameter(const std::string &name, const std::string &value)
{
    if (!server)
        return 1;

    ns1__setParameterResponse response;
    ns1__ParValue parameter;
    parameter.name = name;
    parameter.value = value;
    server->setParameter(&parameter, response);

    return 0;
}


int Alibava::setParameter(const std::string &name, int value)
{
    std::ostringstream ostr;
    ostr << value;
    return setParameter(name, ostr.str());
}
int Alibava::setParameter(const std::string &name, double value)
{
    std::ostringstream ostr;
    ostr << value;
    return setParameter(name, ostr.str());
}


int Alibava::getStatus(ns1__Status &status)
{
    if (!server)
        return 1;

    ns1__getStatusResponse response;
    server->getStatus(0, response);

    ns1__Status *ss = response.status;

    if (ss)
    {
        status.state = ss->state;
        status.nexpected = ss->nexpected;
        status.ntrigger = ss->ntrigger;
        status.rate = ss->rate;
        status.time = ss->time;
        status.value = ss->value;

        return 0;
    }
    else
        return -1;
}

int Alibava::stopRun()
{
    if (!server)
        return 1;

    ns1__stopRunResponse response;
    return server->stopRun(0, response);
}

int Alibava::Reset()
{
    if (!server)
        return 1;

    return server->Reset(0);
}

void Alibava::read_histograms(std::istream &is)
{
    int ihst, nhst;
    const char *titles[] = { "Spectrum", "Channel Profile", 0 };
    reset_histograms();
    is.read((char *)&nhst, sizeof(int));
    for(ihst=0; ihst<nhst;ihst++)
    {
        Histogram *hst = new Histogram(titles[ihst % 2], titles[ihst % 2], 1, 0., 1.);
        hst->import_from_soap(is);

        histograms.push_back(hst);
    }
}

int Alibava::getRunData()
{
    ns1__getRunDataResponse out;
    int rc = server->getRunData(0, out);
    if (out.getRunDataResponse.__size == 0)
        return 1;

    std::istringstream is( std::string((const char *)out.getRunDataResponse.__ptr,
                                               (size_t)out.getRunDataResponse.__size ) );
    read_histograms(is);
    return rc;
}

int Alibava::startRun(ns1__RunType type, bool async, int nevt, int nbin, int xmin, int xmax, int nsample)
{
    ns1__startRunResponse response;
    ns1__DAQParam param;
    int rc = 0;

    if (!server)
        return 1;

    param.nbin = nbin;
    param.xmin = xmin;
    param.xmax = xmax;
    param.numEvents = nevt;
    param.sampleSize = nsample;
    param.runType = type;

    if (!async)
    {
        server->startRun(&param, response);

        std::istringstream is( std::string((const char *)response.startRunResponse.__ptr,
                                           (size_t)response.startRunResponse.__size ) );
        read_histograms(is);
    }
    else
        rc = server->startRunAsync(&param);

    return rc;
}

int Alibava::startPedestalRun(int nevt, bool async, int nbin, int xmin, int xmax, int nsample)
{

    return startRun(ns1__RunType__Pedestal, async, nevt, nbin, xmin, xmax, nsample);

}


int Alibava::startSourceRun(int nevt, bool async, int nbin, int xmin, int xmax, int nsample)
{
    return startRun(ns1__RunType__RadioactiveSource, async, nevt, nbin, xmin, xmax, nsample);
}


int Alibava::startLaserRun(int nevt, bool async, int nbin, int xmin, int xmax, int nsample)
{
    return startRun(ns1__RunType__Laser, async, nevt, nbin, xmin, xmax, nsample);
}

int Alibava::startCalibrationRun(bool async, int nevt_per_point, int npts, double vfrom, double vto, bool is_charge)
{
    int cal_type = is_charge ? 1 : 0;
    return startRun(ns1__RunType__Calibration, async, nevt_per_point, cal_type, npts, vfrom, vto);
}
int Alibava::startChargeScan(bool async, int nevt_per_point, int npts, double vfrom, double vto)
{
    return startRun(ns1__RunType__ChargeScan, async, nevt_per_point, 1, npts, vfrom, vto);
}
int Alibava::startLaserSync(bool async, int nevt_per_point, int npts, double vfrom, double vto)
{
    return startRun(ns1__RunType__Calibration, async, nevt_per_point, 0, npts, vfrom, vto);
}

