/* -*- mode: c++ -*- */
#ifndef __AlibavaConfig_h__
#define __AlibavaConfig_h__

#include <PluginConfig.h>
#include <ScanConfig.h>
#include <BeetleConfig.h>
#include <LaserConfig.h>
#include <TriggerConfig.h>
#include <DAQConfig.h>
#include <Analysis.h>
#include <ComparatorConfig.h>
#include <GladeApp.h>

/**
 * This is the base class for the objects used to set configuration values
 * via the GUI. The only exported method is set_values which should be called
 * when quiting the GUI dialog
 */
class ConfigGUI : public AlibavaGladeApp
{
    private:
        int _firmware;
    public:
        ConfigGUI(int fw) : _firmware(fw)
        {
        }
        virtual ~ConfigGUI()
        {
        }

        virtual void set_values()
        {
        }

        virtual void get_values()
        {
        }

        int firmware() const { return _firmware; }
};

/*
 * Sets the values relative to the calibration scans
 * It has an instance of CalibrationConfig to store
 * the values
 */
class CalibrationGUI : public ConfigGUI
{
    protected:
        CalibrationConfig _cnf;
        bool do_trim_calib;

        static void on_btn_calib_trim(GtkWidget *, CalibrationGUI *);
    public:
        CalibrationGUI(CalibrationConfig &c, int fw) :
            ConfigGUI(fw), _cnf(c), do_trim_calib(false)
        {
        }
        ~CalibrationGUI()
        {
        }

        static void on_scan_changed(GtkWidget *, CalibrationGUI *);

        // GladeApp stuff
        void user_init();
        void process_text(const std::string &s, GtkWidget *w);

        // Returns the local values of Calibration config
        const CalibrationConfig &config() const
        {
            return _cnf;
        }
        void set_values();
        void check_values();
        void auto_connect();
        bool trim_calib() const { return do_trim_calib; }
};

/**
 * Dialog to set up the parameters for the LaserSync mode
 */
class LaserSyncGUI : public ConfigGUI
{
    private:
        LaserSyncConfig _cnf;
    public:
        LaserSyncGUI(LaserSyncConfig &c, int fw) :
            ConfigGUI(fw), _cnf(c)
        {
        }
        ~LaserSyncGUI()
        {
        }

        // GladeApp stuff
        void user_init();
        void set_values();

        const LaserSyncConfig &config() const
        {
            return _cnf;
        }
};

class BeetleConfigGUI : public ConfigGUI
{
    private:
        BeetleConfig _cnf;
    public:
        BeetleConfigGUI(BeetleConfig &c, int fw) :
            ConfigGUI(fw), _cnf(c)
        {
        }
        ~BeetleConfigGUI()
        {
        }

        // GladeApp stuff
        void user_init();
        void set_values();

        const BeetleConfig &config() const
        {
            return _cnf;
        }
};

class TriggerConfigGUI : public ConfigGUI
{
    private:
        TriggerConfig _cnf;
    public:
        TriggerConfigGUI(TriggerConfig &t, int fw) :
            ConfigGUI(fw), _cnf(t)
        {
        }
        ~TriggerConfigGUI()
        {
        }

        // GladeApp stuff
        void user_init();
        void set_values();

        const TriggerConfig &config() const
        {
            return _cnf;
        }

};

class LaserConfigGUI : public ConfigGUI
{
    private:
        LaserConfig _cnf;
    public:
        LaserConfigGUI(LaserConfig &c, int fw) :
            ConfigGUI(fw), _cnf(c)
        {
        }
        ~LaserConfigGUI()
        {
        }
        // GladeApp stuff
        void user_init();
        void set_values();

        const LaserConfig &config() const
        {
            return _cnf;
        }
};

class DAQConfigGUI : public ConfigGUI
{
    private:
        DAQConfig _cnf;
    public:
        DAQConfigGUI(DAQConfig &c, int fw) :
            ConfigGUI(fw), _cnf(c)
        {
        }
        ~DAQConfigGUI()
        {
        }

        void user_init();
        void process_text(const std::string &s, GtkWidget *w);
        void set_values();

        const DAQConfig &config() const
        {
            return _cnf;
        }
};

class AnalysisConfigGUI : public ConfigGUI
{
    private:
        Analysis _cnf;

        static void on_activate_common_mode(GtkWidget *, AnalysisConfigGUI *);
    public:
        AnalysisConfigGUI(Analysis &a, int fw) :
            ConfigGUI(fw), _cnf(a)
        {

        }
        ~AnalysisConfigGUI()
        {

        }

        void auto_connect();
        void user_init();
        void set_values();

        const Analysis &config() const
        {
            return _cnf;
        }
};

class PluginConfigGUI : public ConfigGUI
{
    private:
        PluginConfig _cnf;

        static void on_browse_lib(GtkWidget *, PluginConfigGUI *);
        static void on_browse_path(GtkWidget *, PluginConfigGUI *P);
        static void on_find_symbols(GtkWidget *, PluginConfigGUI *P);
    public:
        PluginConfigGUI(PluginConfig &p, int fw) : ConfigGUI(fw), _cnf(p) {}
        ~PluginConfigGUI() {}


        void auto_connect();
        void user_init();
        void set_values();

        const PluginConfig &config() const
        {
            return _cnf;
        }
};

class ComparatorConfigGUI : public ConfigGUI
{
    private:
        ComparatorConfig _cnf;

        static void on_enable(GtkWidget *, ComparatorConfigGUI *);
        static void btn_mask_all(GtkWidget *, ComparatorConfigGUI *);
        static void btn_clear_all(GtkWidget *, ComparatorConfigGUI *);
        static void btn_reset_trim(GtkWidget *, ComparatorConfigGUI *);
        static void on_new_chip(GtkWidget *w, ComparatorConfigGUI *);
        static void btn_toggled(GtkWidget *, ComparatorConfigGUI *);
        static void txt_changed(GtkWidget *, ComparatorConfigGUI *);
        static void thrs_changed(GtkWidget *, ComparatorConfigGUI *);
        static void icomp_changed(GtkWidget *, ComparatorConfigGUI *);
        static void polarity_changed(GtkWidget *, ComparatorConfigGUI *);

        void draw_table(GtkTable *, int ichip, int type);

    public:
        ComparatorConfigGUI(ComparatorConfig &C, int fw);
        ~ComparatorConfigGUI() {}

        void auto_connect();
        void user_init();
        void set_values();

        void process_text(const std::string &s, GtkWidget *w);
        const ComparatorConfig &config() const { return _cnf; }


};

struct ConfigTrimPriv;
class ComparatorConfigTrimGUI : public ConfigGUI
{
    private:
        ComparatorConfig _cnf;
        ConfigTrimPriv *_priv;
        int  _idle;
        bool _running;

        static void on_button_start(GtkWidget *, ComparatorConfigTrimGUI *);
        static void on_compute_trim(GtkWidget *, ComparatorConfigTrimGUI *);
        static void on_out_folder(GtkWidget *, ComparatorConfigTrimGUI *);
        static void on_save_data(GtkWidget *, ComparatorConfigTrimGUI *);
        static void on_btn_clicked(GtkWidget *, ComparatorConfigTrimGUI *);
        static void on_new_chip(GtkWidget *w, ComparatorConfigTrimGUI *);
        static void polarity_changed(GtkWidget *, ComparatorConfigTrimGUI *);

        static GtkWidget *create_histograms(ComparatorConfigTrimGUI *, char *name, char *s1, char *s2, int i1, int i2);
    public:
        ComparatorConfigTrimGUI(ComparatorConfig &C, int fw);
        ~ComparatorConfigTrimGUI();

        void auto_connect();
        void user_init();
        void set_values();
        void set_handlers();
        const ComparatorConfig &config() const { return _cnf; }

        bool running() const { return _running; }
        void start();
        void stop();

        void set_idle(int v) { _idle=v; }
        int get_idle() const { return _idle; }
        void show_recomended_value(int val);
};

#endif
