/*
 * AlibavaMonitor.cc
 *
 *  Created on: Jul 20, 2012
 *      Author: lacasta
 */

#include <iostream>
#include <iomanip>
#include <sstream>
#include <glib.h>
#include "AlibavaMonitor.h"
#include "AlibavaGUI.h"
#include "utils.h"

AlibavaMonitor __monitor__;

struct AlibavaMonitorPriv
{
        bool ignore_mutex;
        AlibavaMutex mtx;
        AlibavaMonitorPriv() : ignore_mutex(false) {}
};

AlibavaMonitor::AlibavaMonitor()
    : _time_elapsed("0 0:00:00"), _rate("0 Hz"),
      _run_type_val(0), _ntrigger(0), _nexpected(0), _naccepted(0),
      _eff(0.0), _fraction(0.0), _ratef(0.0), _value(0.0), _time(0.0),
      _start_from_soap(false), _ntrig0(0.0), _tim0(0.0), _data(0),
      _running(false)
{
    // TODO Auto-generated constructor stub
    _timer = g_timer_new();
    _priv = new AlibavaMonitorPriv;
}

double AlibavaMonitor::getFraction() const
{
    return _fraction;
}


#define GET_VALUE(type, member) {   \
    type ret;                       \
    _priv->mtx.lock();              \
    ret =  member;                  \
    _priv->mtx.unlock();            \
    return ret;                     \
}

#define SET_VALUE(member, val) {    \
    _priv->mtx.lock();              \
    member = val;                   \
    _priv->mtx.unlock();            \
}
int AlibavaMonitor::getNtrigger() const
{
    GET_VALUE(int, _ntrigger);
}
void AlibavaMonitor::newTrigger()
{
    _priv->mtx.lock();
    _ntrigger++;
    _priv->mtx.unlock();
}

const std::string AlibavaMonitor::getOutFile() const
{
    GET_VALUE(std::string, _out_file);
}

const std::string AlibavaMonitor::getRate() const
{
    GET_VALUE(std::string, _rate);
}

double AlibavaMonitor::getRateAsNumber() const
{
    GET_VALUE(double, _ratef);
}

void AlibavaMonitor::setFraction(double fraction)
{
    SET_VALUE(_fraction, fraction);
}

void AlibavaMonitor::setNtrigger(int ntrigger)
{
    SET_VALUE(_ntrigger, ntrigger);
}

void AlibavaMonitor::setOutFile(const std::string& outFile)
{
    SET_VALUE(_out_file, outFile);
}

void AlibavaMonitor::setRate(const std::string& rate)
{
    SET_VALUE(_rate, rate);
}

void AlibavaMonitor::setRate(double val)
{
    SET_VALUE(_ratef, val);
}

const std::string AlibavaMonitor::getRunType() const
{
    GET_VALUE(std::string, _run_type);
}

int AlibavaMonitor::getNexpected() const
{
    GET_VALUE(int, _nexpected);
}

double AlibavaMonitor::getValue() const
{
    GET_VALUE(double, _value);
}

void AlibavaMonitor::setValue(double value)
{
    SET_VALUE(_value, value);
}

void AlibavaMonitor::setNexpected(int nexpected)
{
    SET_VALUE(_nexpected, nexpected);
}

void AlibavaMonitor::setRunType(const std::string& runType)
{
    SET_VALUE(_run_type, runType);
}

void AlibavaMonitor::setTimeElapsed(const std::string& timeElapsed)
{
    SET_VALUE(_time_elapsed, timeElapsed);
}

void AlibavaMonitor::setData(void* data)
{
    SET_VALUE(_data, data);
}

const std::string AlibavaMonitor::getTimeElapsed() const
{
    GET_VALUE(std::string, _time_elapsed);
}

AlibavaMonitor::~AlibavaMonitor()
{
    // TODO Auto-generated destructor stub
    g_timer_destroy(_timer);
    delete _priv;
}

void AlibavaMonitor::getHistogram(const std::string &whst, size_t &size, char **buffer)
{
    getHistogram(whst, "png", size, buffer);
}

void AlibavaMonitor::getHistogram(const std::string &whst, const std::string &type, size_t &size, char **buffer)
{
    AlibavaGUI *gui = (AlibavaGUI *)_data;
    gui->getHistogram(whst, type, size, buffer);
}


int AlibavaMonitor::getRunTypeVal() const
{
    GET_VALUE(int,  _run_type_val);
}
void AlibavaMonitor::setRunTypeVal( int v )
{
    _priv->mtx.lock();
    _run_type_val = v;
    switch (_run_type_val )
    {
        case 1:
            _run_type = "Calibration";
            break;
        case 2:
            _run_type = "LaserSync";
            break;
        case 3:
            _run_type = "Laser";
            break;
        case 4:
            _run_type = "RadSource";
            break;
        case 5:
            _run_type = "Pedestal";
            break;
        default:
            _run_type = "Unknown";
            break;
    }
    _priv->mtx.unlock();
}

double AlibavaMonitor::getNtrig0() const
{
    GET_VALUE(double,_ntrig0);
}

void AlibavaMonitor::setNtrig0(double ntrig0)
{
    SET_VALUE(_ntrig0, ntrig0);
}

double AlibavaMonitor::getTim0() const
{
    GET_VALUE(double,_tim0);
}

int AlibavaMonitor::getNaccepted() const
{
    GET_VALUE(int, _naccepted);
}

void AlibavaMonitor::setNaccepted(int naccepted)
{
    SET_VALUE(_naccepted, naccepted);
}

void AlibavaMonitor::setTim0(double tim0)
{
    SET_VALUE(_tim0, tim0);
}

double AlibavaMonitor::getEfficiency() const
{
    double eff;
    _priv->mtx.lock();
    eff = _ntrigger > 0.0
            ? 100.0*_naccepted/_ntrigger
            : 0.0;
    _priv->mtx.unlock();
    return eff;
}
std::string AlibavaMonitor::getEfficiencyAsString(int set_prec) const
{
    std::ostringstream ostr;
    double eff = getEfficiency();
    ostr << std::setiosflags(std::ios::left | std::ios::fixed)
         << std::setprecision(set_prec)
         << eff;
    return ostr.str();
}

std::string AlibavaMonitor::computeTimeElapsed()
{
    std::ostringstream ostr;
    std::string ret;
    int tt, seconds, minutes, hours, days;
    double tm;


    _priv->mtx.lock();
    tm = g_timer_elapsed(_timer, NULL);
    tt = (int)tm;
    seconds = tt % 60;
    minutes = (tt / 60) % 60;
    days = tt / 86400;
    hours = (tt / 3600 ) % 3600 - 24*days;

    ostr << days << " " << std::setfill(' ') << std::setw(2) << hours << ":"
            << std::setfill('0') << std::setw(2) << minutes << ":" << std::setfill('0')
            << std::setw(2) << seconds;

    _time_elapsed = ostr.str();
    ret = _time_elapsed;
    _priv->mtx.unlock();
    return ret;
}

double AlibavaMonitor::getTimeElapsedAsDouble() const
{
    double ret=0.0;

    _priv->mtx.lock();
    ret = _time; //g_timer_elapsed(_timer, NULL);
    _priv->mtx.unlock();

    return ret;
}

void AlibavaMonitor::computeRate(int n_received)
{
    _priv->mtx.lock();
    double nrecv = n_received;
    double ntrg = _ntrigger;
    double naccept = _naccepted;
    double eff = ntrg>0.0 ? 100.0*naccept/ntrg : 0.0;
    double ntrig =  nrecv - _ntrig0;
    double tm = g_timer_elapsed(_timer, NULL);
    double tim0 = tm - _tim0;

    _eff = eff;
    if ( tim0 > 0.5)
    {
        double rate = ntrig/tim0;
        std::ostringstream orate;
        orate << std::setiosflags(std::ios::fixed) << std::setw(6) << std::setprecision(1) << rate << " Hz";

        _rate = orate.str();
        _ratef = rate;
        _ntrig0 = nrecv;
        _tim0 = tm;
    }
    _priv->mtx.unlock();
}

void AlibavaMonitor::isRunning(bool v)
{
    _priv->mtx.lock();
    _running = v;

    if (_running)
    {
        g_timer_reset(_timer);
        g_timer_start(_timer);
        _priv->mtx.unlock();
    }
    else
    {
        g_timer_stop(_timer);
        _time = g_timer_elapsed(_timer, NULL);
        _priv->mtx.unlock();
        computeTimeElapsed();
    }

}

void AlibavaMonitor::reset_counters()
{
    _priv->mtx.lock();
    _ntrig0 = 0;
    _tim0 = 0;
    _time = 0;
    _naccepted = 0;
    _ntrigger = 0;
    _priv->mtx.unlock();
}

double AlibavaMonitor::getEff() const
{
    GET_VALUE(double,_eff);
}

void AlibavaMonitor::setEff(double eff)
{
    SET_VALUE(_eff, eff);
}
