/*
 * CalibTrimRun.cc
 *
 *  Created on: Jun 28, 2014
 *      Author: lacasta
 */

#include <iostream>
#include <fstream>
#include <sstream>
#include <iomanip>
#include <cstdio>
#include <cmath>
#include <vector>
#include <ctime>
#include <gtk/gtk.h>
#include "Alibava.h"
#include "ConfigFactory.h"
#include "ScanConfig.h"
#include "ComparatorConfig.h"
#include "BeetleConfig.h"
#include "CalibTrimRun.h"
#include "DAQmanager.h"
#include "GTimer.h"
#include "gtkhst/stats.h"
#include "gtkhst/graph.h"
#include "gtkhst/histogram.h"
#include "gtkhst/histogram2d.h"

#include "AlibavaConfig.h"
#include "AlibavaMonitor.h"
#include "path_utils.h"


class SilentDAQ
{
    private:
        DAQmanager *dm;
    public:
        SilentDAQ()
        {
            dm = DAQmanager::get_the_manager();
            dm->set_silence(true);
        }
        ~SilentDAQ()
        {
            dm->set_silence(false);
        }
};

class ScanEnvelope
{
    private:
        Emitter *emitter;
        BitSet *mask[2];
        ComparatorConfig *cfg;
    public:
        ScanEnvelope(Emitter *E) : emitter(E), cfg(0)
        {
            for (int ic=0; ic<2;ic++)
                mask[ic] = 0;

            g_signal_emit_by_name(emitter, "start", 0, NULL);
        }
        ~ScanEnvelope()
        {
            g_signal_emit_by_name(emitter, "finished", 1, NULL);
            for (int ic=0; ic<2; ic++)
            {
                if (mask[ic])
                {
                    if (cfg)
                        cfg->setMask(ic, *mask[ic]);

                    delete mask[ic];
                }
            }
        }
        void add_bitset(int i, const BitSet &bs)
        {
            if (mask[i])
            {
                delete mask[i];
                mask[i] = 0;
            }
            mask[i] = new BitSet(bs);
        }
        void set_comparator(ComparatorConfig *C)
        {
            cfg = C;
            for (int ichip=0; ichip<2; ichip++)
                add_bitset(ichip, C->getMask(ichip));

        }
        const BitSet *get_bitset(int i) const { return mask[i]; }
};

void gime_time(char *buff, size_t size)
{
    time_t t = time(0);
    strftime(buff, size, "%F_%H.%M.%S", localtime(&t));
}


void save_histogram(GtkWidget *w, const std::string &folder, const char *prefix)
{
    char line[512], sbuf[128];
    gime_time(sbuf, 128);
    sprintf(line,"%s/%s_%s.csv", folder.c_str(), prefix, sbuf);

    std::ofstream ofile(line);
    if (!ofile)
    {
        std::cout << "Could not open " << line << " for saving data" << std::endl;
    }
    GString *str;
    HstChart1D *hst = HST_CHART1D(w);
    if ( HST_IS_FUNCTION2D(hst->func))
        str = hst_function_print(hst->func);
    else
        str = hst_function_print_list(hst->curves);

    ofile.write(str->str, str->len);
    g_string_free(str, TRUE);
    ofile.close();
}

std::ostream &operator<<(std::ostream &os, const LineFitter &F)
{
    os << "offset: "  << F.getOffset() << " slope: " << F.getSlope();
    return os;
}
void LineFitter::copy(const LineFitter &F)
{
    data = F.data;
    offset = F.offset;
    slope = F.slope;
}

int LineFitter::fit()
{
    std::vector< std::pair<double, double> >::iterator ip;
    double vx=0, vy=0, vxx=0, vyy=0, vxy=0, nn=0;
    double denominator = 0.0;

    for (ip=data.begin(); ip!=data.end(); ++ip)
    {
        double a=ip->first, b=ip->second;
        //printf("(%8.4f, %8.4f) ", a, b);
        if (a<=0)
            continue;
        nn++;
        vx += a;
        vy += b;
        vxx += a*a;
        vyy += b*b;
        vxy += a*b;
    }
    if (nn<=1)
    {
        std::cout << " Not enough points" << std::endl;
        return 1;
    }
    vx/=nn;
    vy/=nn;
    vxx/=nn;
    vyy/=nn;
    vxy/=nn;
    denominator = vxx - vx*vx;
    if ( denominator == 0.0 )
    {
        slope = 0.0;
        printf ("Denominator 0\nPoints follow:\n");
        for (ip=data.begin(); ip!=data.end(); ++ip)
        {
            double a=ip->first, b=ip->second;
            printf("(%8.4f, %8.4f)\n", a, b);
        }
    }
    else
        slope = (vxy-vx*vy)/(vxx-vx*vx);
    offset = vy - slope*vx;
    //printf("y=%8.3f + trim*%8.3f\n", offset, slope);
    return 0;
}


int start_source_run(int nevts, double *T=0)
{
    DAQConfig *daq = dynamic_cast<DAQConfig *>(ConfigFactory::instance()->get("daq"));
    Alibava *alibava = DAQmanager::get_the_manager()->get_board();
    int rc, ntrg=0, nsize = daq->sample_size();
    AlibavaTimer tm;
    double delay = alibava->delay();

    tm.start();
    alibava->stop(true);
    alibava->delay(daq->delay());
    alibava->stop(false);
    rc = alibava->rad_source_acq(nevts);
    tm.stop();

    if (rc==Alibava::NoProblem)
    {
        int nread = alibava->rad_source_reading(false, 0);
        ntrg = alibava->get_trigger_counter();
    }
    if (T)
        *T = tm();

    if (rc<0)
    {
        if (rc<0)
            ntrg = -2;

        if (rc==Alibava::TimeOut)
            ntrg = 0;

        alibava->software_reset();
        ConfigFactory::instance()->config_board(*alibava);
    }

    return ntrg;
}

CalibTrimRun::CalibTrimRun(Alibava *B, ComparatorConfig *cnf, int target, int nevts, int mask, GtkWidget *gain, GtkWidget *trim)
    : _nevts(2*nevts), _mask(mask), _thrs(target), _polarity(0),
      hGain(gain), hTrim(trim), board(B), _stop_run(false),
      do_show_spread(true), do_fast_scan(true), do_trim_scan(false), do_find_thrs(true), do_comp_gain(false), do_current(false), do_thrs_scan(false),
      do_save_data(false), do_fix_trim(false), fix_trim_value(20), comp(cnf), gui(0)
{
    B->set_data_receiver(this);
    if (_mask != B->mask() )
        _mask = B->mask();

    int nstep = _thrs/1024;
    int xtra  = _thrs % 1024;
    if (xtra>512)
        nstep++;

    _thrs = nstep*1024;

    emitter = emitter_new();


    // The trim fits...
    int ichan0, ichan1, ichip0, ichip1;
    find_channel_interval(ichan0, ichan1, ichip0, ichip1);
    for (int ichan=ichan0; ichan<=ichan1; ichan++)
    {
        double slope, offset;
        comp->getTrimLine(ichan/128, ichan%128, offset, slope);
        LineFitter fitter(offset, slope);
        trim_fits.insert( ChanFitter::value_type(ichan, fitter) );
    }
}



CalibTrimRun::~CalibTrimRun()
{
    g_object_unref(emitter);
    board->remove_data_receiver(this);
}


void CalibTrimRun::todo_list(bool show_spread, bool fast_scan, bool do_trim, bool find_thrs, bool comp_gain, bool current, bool scan)
{
    do_show_spread = show_spread;
    do_fast_scan = fast_scan;
    do_trim_scan = do_trim;
    do_find_thrs = find_thrs;
    do_comp_gain = comp_gain;
    do_current = current;
    do_thrs_scan = scan;
}

void CalibTrimRun::define_scan(int nevts, int npts, double from, double to)
{
    trg_count.clear();
}
void CalibTrimRun::end_of_scan_point(double value, unsigned int ntrg)
{
    //std::cout << "Val " << value << " ntrg " << ntrg << std::endl;
    trg_count.push_back( PointVal(value, ntrg) );
}

void CalibTrimRun::start_calib()
{
    /*
     * With the Envelope object we are sure that we will emit a signal
     * when starting and another one when quitting.
     */
    ScanEnvelope envelope(getEmitter());
    BeetleConfig *beetle = dynamic_cast<BeetleConfig *>(ConfigFactory::instance()->get("beetle"));

    envelope.set_comparator(&beetle->get_comparator_settings());

    bool folder_exists = PathUtils::exists(folder);
    bool save_scans = false;
    if (do_save_data)
    {
        if (!folder_exists)
        {
            std::string ofolder("<empty>");
            if ( !folder.empty() )
                ofolder = folder;
            GtkDialogFlags flags = (GtkDialogFlags)GTK_DIALOG_DESTROY_WITH_PARENT;
            GtkWidget * dialog = gtk_message_dialog_new (0,
                                                         flags,
                                                         GTK_MESSAGE_ERROR,
                                                         GTK_BUTTONS_CLOSE,
                                                         "Output folder %s does not exist.",
                                                         ofolder.c_str());
            gtk_dialog_run (GTK_DIALOG (dialog));
            gtk_widget_destroy (dialog);
            return;
        }
        save_scans = true;
    }


    comp->setEnable(true);
    // Load comparator settings
    trim_fits.clear();
    gain_fits.clear();
    /*
     * Find the channel interval from the value of the chip mask
     */
    int ichan0, ichan1, ichip0, ichip1;
    find_channel_interval(ichan0, ichan1, ichip0, ichip1);
    for (int ichan=ichan0; ichan<=ichan1; ichan++)
    {
    	//int ic = ichan % 128;
    	int ichip = ichan/128;
    	double slope, offset;
    	comp->getTrimLine(ichip,ichan,offset,slope);
    	trim_fits[ichan] = LineFitter(offset, slope);
    }

    for (int ichip=ichip0; ichip<=ichip1; ichip++)
        comp->setIcomp(ichip, icomp());

    beetle->set_comparator_settings(*comp);
    ConfigFactory::instance()->config_beetle_comparator(*board);
    _stop_run = false;

    if ( do_show_spread )
    {
        show_spread(do_fix_trim);
        if ( save_scans )
            save_histogram(hGain,folder, "Threshold_spread_0");

        if (_stop_run)
        {
            _stop_run = false;
            return;
        }
    }

    if ( do_fast_scan )
    {
        fast_calib();

        if (_stop_run)
        {
            _stop_run = false;
            return;
        }
    }
    else if ( do_trim_scan )
    {
        find_trim_dacs();
        if (_stop_run)
        {
            _stop_run = false;
            return;
        }
    }

    if (do_find_thrs )
    {
        find_threshold();
        if (_stop_run)
        {
            _stop_run = false;
            return;
        }


        if (do_show_spread)
        {
            show_spread(false);
            if (_stop_run)
            {
                _stop_run = false;
                return;
            }
        }
    }

    if (do_comp_gain)
    {
        find_comparator_gain();
        if (_stop_run)
            _stop_run = false;
    }

    if ( do_current )
    {
        find_current_setting();
        if (_stop_run)
        {
            _stop_run = false;
            return;
        }
    }

    if (do_thrs_scan)
    {
        threshold_scan();
        if (_stop_run)
        {
            _stop_run = false;
            return;
        }
    }

    beetle->set_comparator_settings(*comp);
}

void CalibTrimRun::stop()
{
    _stop_run = true;
}

/*
 * Tries to find at which threshold there is no trigger. Assumes
 * the input is the noise. It will return 255 if no threshold is found for
 * which the number of triggers is 0
 */
int CalibTrimRun::find_zero_counts_threshold(int ichip, int thrs, int step, int mx_evts, gpointer hst)
{
    int izero = 0;
    int thrs0 = thrs;
    double T=0;
    HstHistogram *H = 0;

    if (hst)
        H = HST_HISTOGRAM(hst);



    for (int ith = thrs0; ith < 255; ith += step)
    {
        int ntrg;
        if (_stop_run)
            break;

        board->beetle_config_mainth(ichip, ith);
        ntrg = start_source_run(mx_evts, &T);
        if (ntrg == 0)
        {
            // If izero==0, is the first time we find 0 triggers
            // We try once more. We will return this value
            if (!izero)
            {
                thrs = ith;
            }
            izero++;
            if (izero > 1)
                break;
        }
        else
        {
            izero = 0;
            thrs = 255;
        }
        if (H)
            hst_histogram_fill(H, ith, ntrg);

        std::cout << ith << " -> " << ntrg << " time "  << T << std::endl;
    }
    return thrs;
}



void CalibTrimRun::fast_calib()
{
    SilentDAQ sdaq;
    ConfigFactory *cfg = ConfigFactory::instance();
    BeetleConfig *beetle = dynamic_cast<BeetleConfig *>(cfg->get("beetle"));

    /*
     * We trigger on noise----
     * The idea is to scan the main threshold until we get zero triggers.
     * Then we start scanning the trims to find their 50%
     */
    DAQmanager::get_the_manager()->reset();

    /*
     * Find the channel interval from the value of the chip mask
     */
    int nchan=0, ichan0, ichan1, ichip0, ichip1;
    find_channel_interval(ichan0, ichan1, ichip0, ichip1);
    nchan = ichan1-ichan0 +1;

    /*
     * Create the histograms
     */

    HstHistogram *hst = hst_histogram_new(256, 0.0, 256.0, NULL);
    HstHistogram *hCorr = hst_histogram_new(nchan, ichan0-0.5, ichan1+0.5, NULL);

    /*
     * Program initial values for channel mask and trim
     * Initial trim is 0, which is higher threshold.
     */
    for (int ichan=ichan0; ichan<=ichan1; ichan++, nchan++)
    {
        int ichip = ichan/128;
        int ch = ichan % 128;
        comp->setMask(ichip, ch, 1);
        comp->setTrim(ichip, ch, 0);
    }
    for (int ichip=ichip0; ichip<=ichip1; ichip++)
    {
        board->beetle_config_trim(ichip, comp->getTrim(ichip));
        board->beetle_config_mask(ichip, comp->getMask(ichip));
    }


    // We set the DAQ delay to something small so that we do not wait for too long---
    DAQConfig *daq = dynamic_cast<DAQConfig *>(ConfigFactory::instance()->get("daq"));
    int daq_delay = daq->delay();
    int daq_nsamples = daq->sample_size();
    int daq_nevts = daq->nevts();
    int step = 10;
    daq->delay(0.75);
    daq->nevts(_nevts);
    daq->sample_size(daq->nevts());

    for (int ichip = ichip0; ichip<=ichip1; ichip++)
    {
        int thrs = 0;
        ichan0 = ichip*128;
        ichan1 = ichan0 + 128;

        // Unmask the channels
        for (int ichan=ichan0; ichan<=ichan0+1;ichan++)
            comp->setMask(ichip, ichan % 128, 0);

        board->beetle_config_mask(ichip, comp->getMask(ichip));

        g_object_set(G_OBJECT(hGain), "func", hst,
                                      "xtitle", "Threshold",
                                      "ytitle", "N. Events",
                                      "hymin", 0.0, "hymax", 1.2*_nevts,
                                      NULL);

        // coarse search
        thrs = find_zero_counts_threshold(ichip, 0, step, _nevts, hst);

        if (_stop_run)
            return;

        // detailed search
        thrs = find_zero_counts_threshold(ichip, thrs-step, 1, _nevts, hst);
        thrs -= 20;
        if (thrs<0)
            thrs=0;

        if (_stop_run)
            return;

        if (gui)
        	gui->show_recomended_value(thrs);

        g_object_set(G_OBJECT(hGain), "func", hCorr,
                                      "xtitle", "Channel",
                                      "ytitle", "Threshold Correction",
                                      "hymin", 0.0, "hymax", 32.0,
                                      NULL);

        // we set the threshold
        std::cout << "Final threshold is " << thrs << std::endl;
        comp->setMainTh(ichip, thrs);
        board->beetle_config_mainth(ichip, comp->getMainTh(ichip));

        // Mask the channels
        for (int ichan=ichan0; ichan<=ichan1;ichan++)
            comp->setMask(ichip, ichan % 128, 1);
        board->beetle_config_mask(ichip, comp->getMask(ichip));

        /*
         * Now find the 50% by scanning the threshold corrections
         */

        for (int ichan=ichan0; ichan<=ichan1;ichan++)
        {
            std::ostringstream ostr;
            ostr << "Channel " << ichan;

            // Activate the current channel
            comp->setMask(ichip, ichan % 128, 0);
            board->beetle_config_mask(ichip, comp->getMask(ichip));


            HstHistogram *hst2d = hst_histogram_new(32, 0, 32, 0);
            g_object_set(G_OBJECT(hTrim), "func", hst2d,
                                          "xtitle", "Channel",
                                          "ytitle", ostr.str().c_str(),
                                          NULL);

            // now we scan the trim corrections
            int ntrg;
            int izero=0;
            int trim=0;
            double T=0;
            for (int itrim=0; itrim<32; itrim++)
            {
                if (_stop_run)
                    return;

                comp->setTrim(ichip, ichan, itrim);
                board->beetle_config_trim(ichip, comp->getTrim(ichip));


                ntrg = start_source_run(_nevts, &T);
                hst_histogram_fill(hst2d, itrim, ntrg);
                std::cout << "Chan " << ichan << " itrim " << itrim << " ntrg " << ntrg << " T: " << T << std::endl;

                if (ntrg <= 0)
                {
                    if (!izero)
                        trim = itrim;

                    izero++;
                    if (izero >=1 )
                        break;
                }
                else
                {
                    izero = 0;
                    trim = 0;
                }
            }

            // Store the trim for this channel
            std::cout << "Channel " << ichan << " Trim is " << trim << std::endl;
            comp->setTrim(ichip, ichan % 128, trim);
            hst_histogram_fill(hCorr, ichan, trim);

            // mask again the channel
            comp->setMask(ichip, ichan % 128, 1);
            board->beetle_config_mask(ichip, comp->getMask(ichip));
        }

        // Now we increment the main threshold by the step value
        comp->setMainTh(ichip, thrs+step);
        board->beetle_config_mainth(ichip, comp->getMainTh(ichip));
    }
    daq->delay(daq_delay);
    daq->nevts(daq_nevts);
    daq->sample_size(daq_nsamples);
}

void CalibTrimRun::find_scan_limits(int vfrom, int ichip, int step, int &istart, int &iend)
{
    int rc;
    istart = 0;
    iend   = 0;
    int icase = 0, itry=0;
    double fact = 2.0/nevts();
    double W=0.0;

    for (int ith=0; ith<255; ith+=step)
    {
        if (_stop_run)
            break;

        rc = board->beetle_config_mainth(ichip, ith);
        if ( rc == Alibava::TimeOut )
        {
            board->software_reset();
        }
        rc = board->charge_scan_run(85, nevts(), 1, vfrom, vfrom+1);
        if ( rc == Alibava::TimeOut )
        {
            board->software_reset();
        }
        int ntrg = trg_count[0].second;
        W = ntrg * fact;

        if ( W>0.00 && icase==0)
        {
            icase |=1;
            istart = ith;
        }
        if ( W>1.75 )
            icase |= 2;

        if ( (icase & 6)==2 && W==0 )
            icase |= 4;

        if ( comp->getMode()==ComparatorConfig::trackMode && (icase & 2))
        {
            itry++;
            iend = ith;
            if (itry>4)
                break;
        }

        if ( comp->getMode()==ComparatorConfig::pulseMode && (icase & 6)==6)
        {
            itry++;
            iend = ith;
            if (itry>2)
                break;
        }
    }
    if ( istart * iend )
    {
        istart = std::max(0, istart-step);
        iend   = std::min(255, iend+step);
    }
    //std::cout << "istart " << istart << " iend " << iend << std::endl;
}

void CalibTrimRun::find_channel_interval(int &ichan0, int &ichan1, int &ichip0, int &ichip1)
{
    ichan0 = 0, ichan1=0;
    ichip0 = 0, ichip1=1;
    switch(_mask)
    {
        case 1:
            ichan0 = 0;
            ichan1 = 127;
            ichip0 = 0;
            ichip1 = 0;
            break;
        case 2:
            ichan0 = 128;
            ichan1 = 255;
            ichip0 = 1;
            ichip1 = 1;
            break;
        default:
            ichan0=0;
            ichan1=255;
            ichip0=0;
            ichip1=1;
            break;
    }
}

void CalibTrimRun::show_spread(bool fixed_trims)
{
    const int thrstep = 8;
    const int nthrs = 256/thrstep;
    HstHistogram2D *hst;
    double fact = 2.0/nevts();

    /*
     * Find the channel interval from the value of the chip mask
     */
    int nchan=0, ichan0, ichan1, ichip0, ichip1;
    find_channel_interval(ichan0, ichan1, ichip0, ichip1);

    /*
     * Program initial values for channel mask and trim
     */
    for (int ichan=ichan0; ichan<=ichan1; ichan++, nchan++)
    {
        int ichip = ichan/128;
        int ch = ichan % 128;
        comp->setMask(ichip, ch, 1);
        if (fixed_trims)
            comp->setTrim(ichip, ch, fix_trim_value);
    }
    for (int ichip=ichip0; ichip<=ichip1; ichip++)
    {
        board->beetle_config_trim(ichip, comp->getTrim(ichip));
        board->beetle_config_mask(ichip, comp->getMask(ichip));
    }

    /*
     * Create the histogram
     */
    hst = hst_histogram2d_new(nchan, ichan0, ichan1, nthrs, 0, 256, 0);
    g_object_set(G_OBJECT(hGain), "func", hst, "xtitle", "Channel", NULL);

    /*
     * Start the loop on the channels
     */
    for (int ichan=ichan0; ichan<=ichan1; ichan++)
    {
        char htitle[128];
        int istart, iend;
        int ichip = ichan/128;

        if (_stop_run)
            break;

        // Activate the current channel
        comp->setMask(ichip, ichan % 128, 0);
        board->beetle_config_mask(ichip, comp->getMask(ichip));

        /*
         * Find an estimate, with a coarser scan, of the threshold interval
         */
        find_scan_limits(_thrs, ichip, 2*thrstep, istart, iend);
        if (istart==0 && iend==0)
            iend = 255;


        /*
         * make the threshold scan on the interval found above
         */
        HstVal::iterator ip;
        for (int ith=istart; ith<iend; ith+=thrstep)
        {
            sprintf(htitle,"Chan. %3d - Thrs. %3d", ichan, ith);
            g_object_set(G_OBJECT(hGain), "ytitle", htitle, NULL);

            if (_stop_run)
                break;

            // Set the threshold
            board->beetle_config_mainth(ichip, ith);

            // Do the "run"
            board->charge_scan_run(85, nevts(), 1, _thrs, _thrs+1024);

            // Fill the histogram
            double val = trg_count[0].second;
            double W = fact*val;
            hst_histogram2d_fill(hst, ichan, ith, W);
        }

        // Mask again the channel
        comp->setMask(ichip, ichan % 128, 1);
    }
    g_object_set(G_OBJECT(hGain), "ytitle", "", NULL);
}

struct ScurveData
{
	double thrs;
	double W;
};

void CalibTrimRun::find_comparator_gain()
{
    HstHistogram2D *hst;
    double fact = 2.0/nevts();
    int ichmin = _thrs - 2*3072;
    if (ichmin < 3072)
    	ichmin = 3072;

    std::vector<int> charge;
    for (int ic=0; ic<5; ic++)
    {
        int val = ichmin + ic*3072;
        charge.push_back(val);
    }
    int npts = charge.size();
    double step = (charge[1]-charge[0])/2.0;
    hst = hst_histogram2d_new(npts,
                              charge[0]-step, charge[4]+step,
                              128, 0, 256, 0);
    g_object_set(G_OBJECT(hGain), "func", hst, "xtitle", "Charge (e-)", "ytitle", "Threshold", NULL);

    int ichan0, ichan1, ichip0, ichip1;
    find_channel_interval(ichan0, ichan1, ichip0, ichip1);
    for (int ichan=ichan0; ichan<=ichan1; ichan++)
    {
        int ichip = ichan/128;
        int ch = ichan % 128;
        //comp.setTrim(ichip, ch, 21);
        comp->setMask(ichip, ch, 1);
        if ( do_fix_trim )
            comp->setTrim(ichip, ch, fix_trim_value);
    }
    for (int ichip=ichip0; ichip<=ichip1; ichip++)
    {
        board->beetle_config_trim(ichip, comp->getTrim(ichip));
        board->beetle_config_mask(ichip, comp->getMask(ichip));
    }

    gain_fits.clear();
    for (int ichan=ichan0; ichan<=ichan1; ichan++)
    {
        int istart, iend;
        int ichip = ichan/128;
        char htitle[128];

        if (_stop_run)
            break;



        comp->setMask(ichip, ichan % 128, 0);
        board->beetle_config_mask(ichip, comp->getMask(ichip));

        find_scan_limits(charge[npts-1],ichip, 16, istart, iend);

        HstVal::iterator ip;
        hst_histogram2d_reset( hst );
        sprintf(htitle,"Threshold - Chan. %d", ichan);
        g_object_set(G_OBJECT(hGain), "ytitle", htitle, NULL);
        gtk_widget_queue_draw(GTK_WIDGET(hGain));

        ScurveData Walues[2][npts];
        for (int ipoint=0; ipoint<npts; ipoint++)
        {
        	Walues[0][ipoint].thrs = Walues[1][ipoint].thrs = -1;
        	Walues[0][ipoint].W = Walues[1][ipoint].W = 0.0;
        }
        for (int ith=istart; ith<iend; ith+=2)
        {
            if (_stop_run)
                break;
            board->beetle_config_mainth(ichip, ith);
            board->charge_scan_run(85, nevts(), npts, charge[0], charge[npts-1]);
            for (ip=trg_count.begin(); ip!=trg_count.end(); ++ip)
            {
                double Q = ip->first;
                double W = fact*ip->second;
                int ibin = hst_function_find_bin(HST_FUNCTION(hst), Q);
                if (W<0.5)
                {
                	Walues[0][ibin].thrs = ith;
                	Walues[0][ibin].W = W;
                }
                if (W>0.5 && Walues[1][ibin].thrs < 0)
                {
                	Walues[1][ibin].thrs = ith;
                	Walues[1][ibin].W = W;
                }
                hst_histogram2d_fill(hst, Q, ith, W);
            }
        }
        /*
         * Now we fit the curve
         */
        LineFitter fitter;
        for (int ipoint=0; ipoint<npts; ipoint++)
        {
        	double V50p;
        	double D  = Walues[1][ipoint].thrs;
        	double D0 = Walues[0][ipoint].thrs;
        	double W  = Walues[1][ipoint].W;
        	double W0 = Walues[0][ipoint].W;
        	double QQ = hst_function_bin_center(HST_FUNCTION(hst), ipoint);
        	double delta = W - W0;
        	if (delta != 0.0)
        	{
        		V50p = (D-D0)*(0.5-W0)/(W-W0)+D0;
        		fitter.add(QQ, V50p);
        	}
        }
        if (! fitter.fit() )
        {
        	//std::cout << "Channel " << ichan << " Fit: " << fitter << std::endl;
        	comp->set_gain(ichip, ichan % 128, fitter.getSlope());
        }
        else
        {
        	comp->set_gain(ichip, ichan % 128, 0.0);
        }
        gain_fits.insert( ChanFitter::value_type(ichan, fitter) );

        comp->setMask(ichip, ichan % 128, 1);
        if (do_save_data)
        {
            std::ostringstream ostr;
            ostr << "comp_gain_chan_" << std::setw(3) << std::setfill('0') << ichan;
            save_histogram(hGain,folder, ostr.str().c_str());
        }
    }
}
void CalibTrimRun::find_trim_dacs()
{
    char name[128];
    HstHistogram2D *hst;
    double fact = 2.0/nevts();
    const int ntrim =  4,   trimstep=32/ntrim;
    const int nth   = 64, thstep = 256/nth;
    hst = hst_histogram2d_new(ntrim, 0, 32, nth, 0, 256, 0);
    g_object_set(G_OBJECT(hTrim), "func", hst, NULL);

    int ichan0, ichan1, ichip0, ichip1;
    find_channel_interval(ichan0, ichan1, ichip0, ichip1);
    for (int ichan=ichan0; ichan<=ichan1; ichan++)
    {
        int ichip = ichan/128;
        int ch = ichan % 128;
        comp->setTrim(ichip, ch, 0);
        comp->setMask(ichip, ch, 1);
    }
    for (int ichip=ichip0; ichip<=ichip1; ichip++)
    {
        board->beetle_config_mainth(ichip, _thrs);
        board->beetle_config_mask(ichip, comp->getMask(ichip));
    }

    trim_fits.clear();
    for (int ichan=ichan0; ichan<=ichan1; ichan++)
    {
        LineFitter fitter;
        int ichip = ichan/128;
        int val;
        if (_stop_run)
            break;

        /*
         * Clear the histogram and set the name with the channel number
         */
        hst_histogram2d_reset( hst );
        sprintf(name, "Ithmain - Channel %d", ichan);
        g_object_set(G_OBJECT(hTrim),"ytitle", name, "xtitle", "Trim Corr.", NULL);

        /*
         * Unmask the channel
         */
        comp->setMask(ichip, ichan % 128, 0);
        board->beetle_config_mask(ichip, comp->getMask(ichip));

        /*
         * Loop over trim values
         */
        int ith0 = 0;
        int ival = 0;
        //std::cout << "Channel " << ichan << std::endl;
        for (int itrim=trimstep/2; itrim<32; itrim += trimstep, ival++)
        {
            int istart=0;
            int istop=0;
            int icase=0;
            double W, W0=-1, V50p;
            double D, D0;
            if (_stop_run)
                break;

            /**
             * TODO: add a flag that allows to scan the whole threshold interval
             *       to appreciate the S-curves
             */
            //std::cout << "Trim " << itrim <<  " [" << ith0 << "]" << std::endl;
            comp->setTrim(ichip, ichan % 128, itrim);
            board->beetle_config_trim(ichip, comp->getTrim(ichip));
            for (int ith=ith0; ith<256; ith+=thstep)
            {
                for (int itry=0; itry<20; itry++)
                {
                	if (_stop_run)
                		break;

                	//std::cout << "Ith " << ith << "[" << ith0 << "]" << std::endl;
                	board->beetle_config_mainth(ichip, ith);
                	board->charge_scan_run(85, nevts(), 1, _thrs, _thrs+1024);
                	val = trg_count[0].second;
                	if ( !( val==0 && istart ) )
                		break;
                }
                W = val*fact;
                D = ith;

				// Starting the curve
                if ( W>0.00 && icase==0)
                {
                    icase |=1;
                }
                
                // Well over the read band
                if ( (icase & 7)==3 && W>0 )
                {
                    if (W>0.5)
                    {
                    	W0 = W;
                    	D0 = D;
                    }
                    else if ( !(icase&8) )
                    {
                        V50p = (D-D0)*(0.5-W0)/(W-W0)+D0;
//                        std::cout << "\t*** V50p: thrs " << ith << " val=" << V50p << std::endl
//                                << "\t\t (0.25-" << W0 << ")*(" << D << "-" << D0 << ")/(" << W << "-" << W0 << ") + " << D0 << std::endl;
                        fitter.add(itrim, V50p);
                        icase |= 8;
                    }

                }
                
                // Past the red stripe
                if ( W>1.75 && !(icase & 2)  ) {
                    icase |= 2;
                    if ( comp->getMode()==ComparatorConfig::pulseMode )
                        ith0 = ith;
                }
                
                if ( comp->getMode()==ComparatorConfig::trackMode )
                {
					if ( (icase & 2) && W>0.25 )
                    	ith0 = ith;
				}
				// Zero content at the other side of the read stripe.
				// We are at the end
				if ( (icase & 6)==2 && W==0 )
                    icase |= 4;


                if ( (icase & 4) )
                {
                    istop++;
                    if (istop>4)
                        break;
                }

//                if ( comp->getMode()==ComparatorConfig::trackMode && (icase & 2))
//                {
//                    istop++;
//                    if (istop>4)
//                        break;
//                }
//
//                if ( comp->getMode()==ComparatorConfig::pulseMode && (icase & 6)==6)
//                {
//                    istop++;
//                    if (istop>2)
//                        break;
//                }



                hst_histogram2d_fill(hst, itrim, ith, W);

            }
        }
        comp->setMask(ichip, ichan % 128, 1);
        comp->setTrim(ichip, ichan % 128, 0);

        /*
         * Compute the slope and offset
         */
        fitter.fit();
        trim_fits.insert( ChanFitter::value_type(ichan, fitter) );
        if ( do_save_data )
        {
            std::ostringstream ostr;
            ostr << "trim_chan_" << std::setw(3) << std::setfill('0') << ichan;
            save_histogram(hTrim, folder, ostr.str().c_str());
        }
    }
    if (!_stop_run)
        find_threshold();
}

double CalibTrimRun::compute_trim_corr(int ithmain, bool save)
{
    if (trim_fits.empty())
        return -1.;

    int ichan0, ichan1, ichip0, ichip1;
    find_channel_interval(ichan0, ichan1, ichip0, ichip1);
    ConfigFactory *cfg = ConfigFactory::instance();
    BeetleConfig *beetle = dynamic_cast<BeetleConfig *>(cfg->get("beetle"));
    int indx = 0;
    double vtrim;
    HstStats *stats = hst_stats_new();
    for (int ichan=ichan0; ichan<=ichan1; ichan++, indx++)
    {
        LineFitter &F = trim_fits[indx];
        int ichip = ichan/128;
        int ch = ichan % 128;
        vtrim = F.find(ithmain);
        if (vtrim<0)
            vtrim=0;
        if (vtrim>31)
            vtrim=31;

        hst_stats_add(stats, F.eval(vtrim), 0.0, 1.0);
        if (save)
        {
            double offs = F.getOffset();
            double slope = F.getSlope();
            comp->setTrim(ichip, ch, (unsigned char)vtrim);
            comp->setTrimLine(ichip, ch, offs, slope);
        }
    }
    double rms = hst_stats_get_stdev(stats, 0);
    if (save)
    {
        for (int ichip=ichip0; ichip<=ichip1; ichip++)
            comp->setMainTh(ichip, (unsigned short int)ithmain);

        beetle->set_comparator_settings(*comp);
        cfg->config_beetle_comparator(*board);
    }
    return rms;
}

double CalibTrimRun::find_threshold()
{
	int ichan0, ichan1, ichip0, ichip1;
	find_channel_interval(ichan0, ichan1, ichip0, ichip1);

	HstHistogram2D *hst = hst_histogram2d_new(256, 0, 256, 64, 0, 256, 0);
    g_object_set(G_OBJECT(hTrim), "func", hst,
                 "ytitle", "Threshold",
                 "xtitle", "Ithmain",
                 NULL);

    double mean_offset = 0.0;
    double mean_slope = 0.0;
    double xpts = 0.0;
    for (int ichan=ichan0; ichan<=ichan1; ichan++)
    {
        LineFitter &F = trim_fits[ichan];
        mean_offset += F.getOffset();
        mean_slope += F.getSlope();
        xpts ++;
    }
    mean_offset /= xpts;
    mean_slope /= xpts;
    std::cout << " Mean (offset, slope): (" << mean_offset << ", " << mean_slope << ")" << std::endl;

    for (int ichan=ichan0; ichan<=ichan1; ichan++)
    {
        if (trim_fits[ichan].npoits()<2)
            std::cout << "Channel " << ichan << " does not have proper threshold corr." << std::endl;
    }


	/*
	 * TODO: add a graph
	 */
	int npts = 256;
	double xthrs = 9999999;
	double rmsmin = 999999;
	double rmsmax = -9999999;
	double *xval = new double [npts];
	double *yval = new double [npts];
	for (int ith=0; ith<npts; ith++)
	{
		int indx=0;
		double vtrim, vthrs, rms;
		HstStats *stats = hst_stats_new();
		for (int ichan=ichan0; ichan<=ichan1; ichan++, indx++)
		{
			LineFitter &F = trim_fits[ichan];
			if (F.npoits()<2)
			    continue;

			vtrim = F.find(ith);
			if (vtrim<0)
				vtrim=0;
			if (vtrim>31)
				vtrim=31;

			vthrs = F.eval(vtrim);
			hst_stats_add(stats, vthrs, 0.0, 1.0);
			hst_histogram2d_fill(hst, ith, vthrs, 1);
		}
		rms = hst_stats_get_stdev(stats, 0);
		if (rmsmin > rms)
		{
			rmsmin = rms;
			xthrs = ith;
		}
		if (rmsmax < rms)
		    rmsmax = rms;

		xval[ith] = ith;
		yval[ith] = rms;
		g_object_unref(stats);
	}
	HstGraph *gr = hst_graph_new(npts, xval, yval);
    g_object_set(G_OBJECT(hGain), "func", gr,
                 "ytitle", "RMS",
                 "xtitle", "Ithmain",
                 "hymin", 0.0, "hymax", 1.2*rmsmax,
                 NULL);
    delete [] xval;
    delete [] yval;


	/*
	 * Now ask the question...
	 */
	std::cout << "**** Suggested value " << xthrs << std::endl;
	if (gui)
	    gui->show_recomended_value(xthrs);


//	char msg[120];
//	sprintf(msg,"Optimal target threshold is %f\nrms: %f", xthrs, rmsmin);
//	GtkWidget *dlg = gtk_dialog_new_with_buttons("Target Threshold",
//			0, (GtkDialogFlags)( GTK_DIALOG_DESTROY_WITH_PARENT | GTK_DIALOG_MODAL),
//			"Accept target Threshold", GTK_RESPONSE_ACCEPT,
//			"Cancel", GTK_RESPONSE_CANCEL,
//			NULL);
//	GtkWidget *content = gtk_dialog_get_content_area(GTK_DIALOG(dlg));
//	GtkWidget *label = gtk_label_new(msg);
//	gtk_container_add (GTK_CONTAINER (content), label);
//	gtk_widget_show_all (dlg);
//
//	int rc = gtk_dialog_run(GTK_DIALOG(dlg));
//	if (rc == GTK_RESPONSE_ACCEPT)
	{
		BeetleConfig *beetle = dynamic_cast<BeetleConfig *>(ConfigFactory::instance()->get("beetle"));
		ConfigFactory *cfg = ConfigFactory::instance();
		double vtrim;
		for (int ichan=ichan0; ichan<=ichan1; ichan++)
		{
			LineFitter &F = trim_fits[ichan];
			int ichip = ichan/128;
			int ch = ichan % 128;
			vtrim = F.find(xthrs);
			if (vtrim<0)
				vtrim=0;
			if (vtrim>31)
				vtrim=31;

			double offs = F.getOffset();
			double slope = F.getSlope();
			comp->setTrim(ichip, ch, (unsigned char)vtrim);
			comp->setTrimLine(ichip, ch, offs, slope);
		}
		for (int ichip=ichip0; ichip<=ichip1; ichip++)
			comp->setMainTh(ichip, (unsigned short int)xthrs);

		beetle->set_comparator_settings(*comp);
		cfg->config_beetle_comparator(*board);
	}

	if (do_save_data)
	    save_histogram(hTrim, folder, "find_threshold");

	return 0.0;
}

void CalibTrimRun::find_current_setting()
{
    BeetleConfig *beetle = dynamic_cast<BeetleConfig *>(ConfigFactory::instance()->get("beetle"));
    double fact = 2.0/nevts();
    int ichan0, ichan1, ichip0, ichip1, nchan;
    find_channel_interval(ichan0, ichan1, ichip0, ichip1);
    nchan = ichan1 -ichan0 + 1;

    int ichmin = 1024;
    std::vector<int> charge;
    for (int ic=0; ic<25; ic++)
    {
        int val = ichmin + ic*1024;
        charge.push_back(val);
    }
    int npts = charge.size();
    double step = (charge[1]-charge[0])/2.0;


    /*
     * Set current settings
     */
    ComparatorConfig conf_0 = *comp;
    beetle->set_comparator_settings(*comp);
    ConfigFactory::instance()->config_beetle_comparator(*board);

    /*
     * Prepare the histograms
     */
    HstHistogram2D *hst = hst_histogram2d_new(nchan, ichan0, ichan1+1,
                                              npts, charge[0]-step, charge[npts-1]+step,
                                              0);
    g_object_set(G_OBJECT(hGain),
                 "func", hst,
                 "xtitle", "Channel",
                 "ytitle", "Charge (e-)",
                 NULL);

    for (int ichan=ichan0; ichan<=ichan1; ichan++)
    {
        int ichip = ichan/128;
        int ch = ichan % 128;
        comp->setMask(ichip, ch, 1);
    }


    HstVal::iterator ip;
    for (int ichan=ichan0; ichan<=ichan1; ichan++)
    {
        int ichip = ichan/128;
        //int ch = ichan % 128;

        if (_stop_run)
            break;

        comp->setMask(ichip, ichan % 128, 0);
        board->beetle_config_mask(ichip, comp->getMask(ichip));

        board->charge_scan_run(85, nevts(), npts, charge[0], charge[npts-1]);
        for (ip=trg_count.begin(); ip!=trg_count.end(); ++ip)
        {
            double Q = ip->first;
            double W = fact*ip->second;
            hst_histogram2d_fill(hst, ichan, Q, W);
        }
        comp->setMask(ichip, ichan % 128, 1);
        gtk_widget_queue_draw(hGain);
    }

    if (do_save_data)
        save_histogram(hGain, folder, "current_threshold");

    beetle->set_comparator_settings(conf_0);
    ConfigFactory::instance()->config_beetle_comparator(*board);
}

void CalibTrimRun::threshold_scan()
{

}
