/*
 * DAQmanager.cc
 *
 *  Created on: 29 Jan 2011
 *      Author: lacasta
 */

#include <iostream>
#include <sstream>
#include "utils.h"
#include "DAQmanager.h"
#include <RunManager.h>
#include <DataFile.h>
#include <DataMonitor.h>
#include <AsciiDataFile.h>
#include <HDFDataFile.h>
#include "ConfigFactory.h"
#include "Analysis.h"
#include "DAQConfig.h"
#include "PluginConfig.h"
#include "AlibavaMonitor.h"
#include "utils.h"

DAQmanager *DAQmanager::_mngr=0;
AlibavaMutex daq_mutex;

struct DAQmanagerPriv
{
        AlibavaMutex mtx;
};

DAQmanager::DAQmanager(int nn, Alibava *B, RunManager::RunType tp) :
    silence(false), nevts(nn), received(0), expected(nn), board(B), ofile(0), format(DataFile::HDF5), monitor(0),
    run_type(RunManager::Unknown), rm(0),
    _starting(false), _running(false), _stop_running(false), _is_gui(false), config_file( 0), _priv(0)
{
    ConfigFactory *cfg = ConfigFactory::instance();
    config_file = new KeyFile();
    cfg->set_file(config_file);

    DAQConfig *daq = dynamic_cast<DAQConfig *>(cfg->get("daq"));

    daq->nevts(nn);
    board->set_data_receiver(this);
    board->delay( daq->delay() );
    set_run_type(tp);

    _priv = new DAQmanagerPriv;
}

DAQmanager::~DAQmanager()
{
    delete _priv;
    if (rm)
        delete rm;

    daq_mutex.lock();

    if (this == _mngr)
        _mngr = 0;

    daq_mutex.unlock();

    delete config_file;

}


bool DAQmanager::running() const
{
    bool rc;
    _priv->mtx.lock();
    rc = _running;
    _priv->mtx.unlock();
    return rc;
}
void DAQmanager::running(bool x)
{
    _priv->mtx.lock();
    _running = x;
    _starting = false;
    _priv->mtx.unlock();
}


bool DAQmanager::starting() const
{
    bool rc;
    _priv->mtx.lock();
    rc =  _starting;
    _priv->mtx.unlock();
    return rc;
}

bool DAQmanager::stop() const
{
    bool rc;
    _priv->mtx.lock();
    rc =  _stop_running;
    _priv->mtx.unlock();
    return rc;
}
void DAQmanager::stop(bool x)
{
    _priv->mtx.lock();
    _stop_running = x;
    _priv->mtx.unlock();
}

int DAQmanager::open(const char *fnam)
{
    board->open(fnam);
    if (!(*board))
    {
        err_msg( _("Something failed when opening the device") );
        return 1;
    }

    // reset the card
    if (reset()<0)
        return -1;

    device_opened();
    return 0;
}

int DAQmanager::reset()
{
    int rc;

    _priv->mtx.lock();
    if (!is_silent())
        write_msg(_("*** Reseting the board\n"));

    // First thing we do is to send a reset
    rc = board->software_reset();
    if (rc)
    {
        _priv->mtx.unlock();
        return -1;
    }

    rc = ConfigFactory::instance()->config_board(*board);
    _priv->mtx.unlock();
    return rc;
}


void DAQmanager::set_run_type( RunManager::RunType tp )
{
    if (tp == RunManager::Unknown || tp == run_type )
        return;

    if (rm)
        delete rm;

    run_type = tp;
    rm = RunManager::factory(tp, this);
    if (rm)
    {
        ConfigFactory *cfg = ConfigFactory::instance();
        PluginConfig *plg = dynamic_cast<PluginConfig *>(cfg->get("plugin"));
        if ( plg->active())
            rm->set_plugin(plg->plugin());
    }
    if (!rm)
    {
        err_msg( _("Could not create an appropriate RunManager\n") );
    }
}

void DAQmanager::setup_data_file()
{
    _priv->mtx.lock();
    if (!file_name.empty())
    {
        if (ofile)
        {
            write_msg("DAQmanager: This should not happen. Deleting data file.\n");
            delete ofile;
        }
        ofile = DataFile::factory(format, file_name);
    }
    _priv->mtx.unlock();
}

void DAQmanager::set_data_file(const std::string &fnam)
{
    file_name = fnam;
}

int DAQmanager::prepare_for_run()
{
    ConfigFactory *cfg = ConfigFactory::instance();
    DAQConfig *daq = dynamic_cast<DAQConfig *>(cfg->get("daq"));
    PluginConfig *plg = dynamic_cast<PluginConfig *>(cfg->get("plugin"));
    Analysis *ana = dynamic_cast<Analysis *>(cfg->get("analysis"));
    Alibava *alibava = get_board();

    if (!rm)
        return -1;

    if (!is_silent())
        write_msg("Starting a %s run\n", __monitor__.getRunType().c_str());

    received = 0;
    _starting = true;
    board->stop(true);
    board->delay( daq->delay() );
    nevts = daq->nevts( );
    ana->set_mask(  alibava->mask() );
    ana->reset(true);

    if ( plg->active())
        rm->set_plugin(plg->plugin());

    rm->start_of_run();

    if (plg->active())
    {
        int nn = plg->plugin()->start_of_run(run_type, daq->nevts(), daq->sample_size());
        daq->nevts( nn );
    }
    expected = daq->nevts();
    return 0;
}
int DAQmanager::acquire()
{
    ConfigFactory *cfg = ConfigFactory::instance();
    PluginConfig *plg = dynamic_cast<PluginConfig *>(cfg->get("plugin"));
    Analysis *ana = dynamic_cast<Analysis *>(cfg->get("analysis"));
    Alibava *alibava = get_board();

    /*
     * Prepare the output file (if any)
     */
    setup_data_file();
    if (ofile)
    {
        //ofile->scan_definition(0, 0, 0, 0, 0, 0, 0);
        ofile->describe_setup(get_run_type(), alibava->nchip(), alibava->mask(), alibava->firmware());
        ofile->open();
        /*
         * Let the plugin write a header to the file
         */
        if (plg->active())
        {
            std::string header;
            std::string data_format;
            plg->plugin()->get_data_format(data_format);
            ofile->set_data_format(data_format.c_str());
            plg->plugin()->new_file(header);
            if (!header.empty())
                ofile->write(DataFile::NewFile, header.size(), header.c_str());
        }
    }

    std::ostringstream oheader;
    if (ofile)
        ofile->write(run_type,0,0);

    board->stop(false);
    oheader << "V3.0|";

    running(true);
    stop(false);

    int rc = rm->loop(oheader);
    if ( rc >= 0 )
    {
        if ( plg->active() )
        {
            std::string buffer;
            plg->plugin()->end_of_run(buffer);
            if ( ofile )
            {
                if (!buffer.empty())
                    ofile->write(DataFile::EndOfRun, buffer.size(), buffer.c_str());
            }
        }

        if (!is_silent())
            write_msg(_("Run finished\n"));

    }
    else
    {
        reset();
    }

    if (ofile)
    {
        std::ostringstream ostr;
        ana->lock();
        ofile->set_header(oheader.str());
        ofile->write(run_type, ana->pedestals(), ana->noise());
        ana->unlock();
        ofile->close();

        if (!is_silent())
            write_msg(_("Data stored in %s\n"), ofile->file_name().c_str());

        delete ofile;

        ofile = 0;
        file_name.clear();
    }

    // Call the end of run
    rm->end_of_run();

    running(false);
    stop_run();
    return rc;
}

int DAQmanager::stop_run()
{
    stop(true);
    return 0;
}


void DAQmanager::receive(const AlibavaData *b)
{
    _priv->mtx.lock();
    received++;
    _priv->mtx.unlock();

    process_event(b);
}

void DAQmanager::write_event(const AlibavaData *b)
{
    ConfigFactory *cfg = ConfigFactory::instance();
    PluginConfig *plg = dynamic_cast<PluginConfig *>(cfg->get("plugin"));

    if (plg->active())
    {
        std::string data;
        plg->plugin()->filter_event(b, data);
        if (ofile)
        {
            _priv->mtx.lock();
            if (data.empty())
            {
                ofile->write(b);
            }
            else
            {
                ofile->write(DataFile::DataBlock, data.size(), data.c_str());
            }
            _priv->mtx.unlock();
        }
    }
    else
    {
        // write data

        if (ofile)
        {
            _priv->mtx.lock();
            ofile->write(b);
            _priv->mtx.unlock();
        }
    }
}

void DAQmanager::process_event(const AlibavaData *b)
{
    ConfigFactory *cfg = ConfigFactory::instance();
    Analysis *ana = dynamic_cast<Analysis *>(cfg->get("analysis"));
    ana->process_event((unsigned short *)b->chip(0), board->mask());
    write_event(b);
}


void DAQmanager::flush_data_file()
{
    if (ofile)
    {
        _priv->mtx.lock();
        ofile->flush();
        _priv->mtx.unlock();
    }
}
void DAQmanager::load_config(const char *fnam, bool load)
{
    ConfigFactory *cfg = ConfigFactory::instance();

    if (config_file)
        delete config_file;

    config_file = new KeyFile();

    std::ifstream ifile(fnam);
    if (!ifile)
    {
        err_msg( _("could not open config file %s"), fnam);
        return;
    }
    config_file->load(ifile);
    ifile.close();
    cfg->set_file(config_file);

    if (load)
        cfg->config_board( *get_board() );

    config_loaded();
}

void DAQmanager::config_loaded()
{

}

void DAQmanager::save_config(const char *fnam)
{
    ConfigFactory::instance()->save_data();

    std::ofstream cnf_ofile(fnam);
    if (!cnf_ofile)
    {
        err_msg( _("could not dump config: %s"), fnam );
    }
    else
    {
        config_file->save(cnf_ofile);
        cnf_ofile.close();
    }
}

void DAQmanager::set_the_manager(DAQmanager *m)
{
    daq_mutex.lock();
    _mngr=m;
    daq_mutex.unlock();
}

DAQmanager *DAQmanager::get_the_manager()
{
    DAQmanager *out = 0;
    daq_mutex.lock();
    out = _mngr;
    daq_mutex.unlock();

    return out;
}

void DAQmanager::set_daq_manager_data_file(const std::string &fnam)
{
    DAQmanager *dm = get_the_manager();
    if (dm)
    {
        if (dm->running())
            err_msg("Trying to open file while running");

        if (!dm->is_silent())
            write_msg(_("Setting data file to %s"), fnam.c_str());
        dm->set_data_file(fnam);
    }
}

void DAQmanager::daq_manager_stop_run()
{
    DAQmanager *dm = get_the_manager();
    if ( dm )
    {
        if (!dm->is_silent())
            write_msg(_("Stopping run"));

        dm->stop_run();
    }
}
int DAQmanager::daq_manager_reset_board()
{
    int rc = 0;
    DAQmanager *dm = get_the_manager();

    if (dm)
        rc = dm->reset();

    return rc;
}

gboolean update_monitor(gpointer ptr)
{
    if (  DAQmanager::get_the_manager()->is_gui() )
        return FALSE;

    __monitor__.computeTimeElapsed();
    __monitor__.computeRate(DAQmanager::get_the_manager()->get_received());

    return TRUE;
}


void DAQmanager::acquire_thread(bool is_thread)
{
    DAQmanager *dm = get_the_manager();

    dm->prepare_for_run();
    __monitor__.reset_counters();
    __monitor__.isRunning(true);
    dm->acquire();
    stop_run();

    __monitor__.isRunning(false);
//    if (is_thread)
//        g_thread_exit(0);
}

int DAQmanager::daq_manager_start_run(int run_type, int num_events, int sample_size, DataReceiver *D)
{
    ConfigFactory *cfg = ConfigFactory::instance();
    DAQConfig *daq = dynamic_cast<DAQConfig *>(cfg->get("daq"));
    extern AlibavaMonitor __monitor__;
    if ( run_type < 3 )
    {
        daq->nevts(num_events);
        daq->sample_size(sample_size);
    }
    RunManager::RunType the_type;
    int rc;

    DAQmanager *dm = get_the_manager();

    switch (run_type)
    {
        case 0:
            the_type = RunManager::Pedestal;
            break;
        case 1:
            the_type = RunManager::RadSource;
            break;
        case 2:
            the_type = RunManager::Laser;
            break;
        case 3:
            the_type = RunManager::LaserSync;
            break;
        case 4:
            the_type = RunManager::Calibration;
            break;
        case 5:
            the_type = RunManager::ChargeScan;
            break;
        default:
            std::cout << "Unknown Run type" << std::endl;
            return -1;
    }
    __monitor__.setRunTypeVal(the_type);
    __monitor__.setNexpected(dm->get_expected());
    dm->set_run_type(the_type);

    if (D)
        dm->get_board()->set_data_receiver(D);

    __monitor__.from_soap(true);
    int tmout = g_timeout_add(500, (GSourceFunc)update_monitor, 0);
    bool is_thread = false;
    if (dm->is_gui())
    {
        rc = dm->prepare_for_run();
        if (rc)
        {
            if (D)
                dm->get_board()->remove_data_receiver(D);
            return rc;
        }
        is_thread=true;
    }
    dm->acquire_thread(is_thread);

    if (!dm->is_gui())
        g_source_remove(tmout);

    if (D)
        dm->get_board()->remove_data_receiver(D);

    dm->stop_run();

    return 0;
}


int DAQmanager::get_trigger_counter()
{
    if (__monitor__.isRunning())
        return -1;

    return get_the_manager()->get_board()->get_trigger_counter();
}

void DAQmanager::set_expected(int x)
{
    _priv->mtx.lock();
    expected = nevts = x;
    _priv->mtx.unlock();
}
void DAQmanager::set_received(int x)
{
    _priv->mtx.lock();
    received = x;
    _priv->mtx.unlock();
}
int DAQmanager::get_expected() const
{
    int ret;
    _priv->mtx.lock();
    ret = expected;
    _priv->mtx.unlock();
    return ret;
}
int DAQmanager::get_received() const
{
    int ret;
    _priv->mtx.lock();
    ret = received;
    _priv->mtx.unlock();
    return ret;
}
double DAQmanager::get_fraction_received() const
{
    double ret;
    _priv->mtx.lock();
    ret = expected > 0.0 ? double(received) / double(expected) : 0.0;
    _priv->mtx.unlock();
    return ret;
}

