/*
 * HDFDataFile.cc
 *
 *  Created on: Feb 27, 2014
 *      Author: lacasta
 */

#include <ctime>
#include <cstring>
#include <iostream>
#include <vector>
#include <string>
#include <glib.h>
#include <unistd.h>
#include "string_utils.h"
#include "path_utils.h"
#include <HDFDataFile.h>
#include "utils.h"

#ifdef HAVE_CONFIG_H
#ifdef G_OS_WIN32
#include "config_win.h"
#else
#include <config.h>
#endif
#endif

#ifdef HAVE_HDF5
#include <hdf5.h>


typedef struct { char c; short x; } st_short;
typedef struct { char c; int x; } st_int;
typedef struct { char c; long x; } st_long;
typedef struct { char c; float x; } st_float;
typedef struct { char c; double x; } st_double;

#define SHORT_ALIGN (sizeof(st_short) - sizeof(int16_t))
#define INT_ALIGN (sizeof(st_int) - sizeof(int32_t))
#define LONG_ALIGN (sizeof(st_long) - sizeof(int64_t))
#define FLOAT_ALIGN (sizeof(st_float) - sizeof(float))
#define DOUBLE_ALIGN (sizeof(st_double) - sizeof(double))

struct FormatDef
{
        char   fmt;
        char   name[64];
        int    size;   // the type size
        int    len;    // The number of items
        int    align;
        int    offset;
        hid_t  type;
        hid_t  obj;
};
typedef std::vector<FormatDef> ComponentList;


#define SETUP_TYPE(type, ord, NATIVE, LE, BE){  \
    switch (ord)                                \
    {                                           \
        case native:                            \
            type = NATIVE;                      \
            break;                              \
        case little_endian:                     \
            type = LE;                          \
            break;                              \
        case big_endian:                        \
            type = BE;                          \
            break;                              \
    };                                          \
}
typedef std::vector<std::string> TokenList;


class UserData
{
    public:
        enum Ordering { native, little_endian, big_endian };

    private:
        hid_t fileid;
        hid_t data_group;
        size_t data_size;
        unsigned int nevts;
        std::string gname;
        ComponentList components;

        bool parse_data(const char *fmt);
    public:
        UserData(hid_t file_id, const char *group_name, const char *fmt);
        ~UserData();
        bool error() const { return data_group == H5I_BADID; }
        bool write_data(const void *ptr);
        void Close();
};





struct ScanDef
{
        enum ScanTypes { Unknown, Charge, Time, Laser };
        unsigned int type;
        unsigned int from;
        unsigned int to;
        unsigned int npts;
        unsigned int nevts;
        unsigned int charge;
        unsigned int delay;
};

struct Header
{
        int run_type;
        unsigned short nchips;
        unsigned short chip_mask;
        unsigned short firmware;
        float pedestals[256];
        float noise[256];
        struct ScanDef scan;
        struct tm time;
};


struct DataScanPoint
{
        float value; // Value of current scan value
        int   start;    // Event index where this scan point starts
        int   end;      // Event index where this scan point finishes
};

struct AlibavaDataBlock
{
        float temp;
        float time;
        unsigned int   clock;
        unsigned short header[32];
        unsigned short data[256];
};

struct HDFPrivate
{
        Header H; /* The header with the run information */
        DataScanPoint P; /* A data scan Point */
        hid_t fileid;    /* HDF5 identifier for file */
        hid_t header;    /* The header data set */
        hid_t pedID;
        hid_t noiseID;
        hid_t data_group;   /* DAta group */
        hid_t headerID;     /* The DataSet ID for the event headers */
        hid_t dataID;       /* The DataSet ID for data */
        hid_t tempID;       /* */
        hid_t timeID;
        hid_t clockID;
        hid_t scanID;       /* Group for scan points */
        hid_t scvalID;
        hid_t startID;
        hid_t endID;
        hid_t userG;
        hid_t userID;
        unsigned int nevts; /* The number of events writen */
        unsigned int nscan; /* The number of scanpoints writen */
        UserData *user;
        AlibavaMutex mutex;
        HDFPrivate()
            : fileid(H5I_BADID), header(H5I_BADID),
              pedID(H5I_BADID), noiseID(H5I_BADID), data_group(H5I_BADID),
              headerID(H5I_BADID), dataID(H5I_BADID),
              tempID(H5I_BADID), timeID(H5I_BADID), clockID(H5I_BADID),
              scanID(H5I_BADID), scvalID(H5I_BADID), startID(H5I_BADID),endID(H5I_BADID),
              userG(H5I_BADID), userID(H5I_BADID), nevts(0), nscan(0), user(0)
        {
        }
        ~HDFPrivate()
        {
        }
};

/**
 * Adds an attribute
 */
static
void add_attribute(hid_t dest, const char *name, hid_t type, void *buf)
{
    hsize_t adim[1]={1}, admx[1]={1};
    hid_t aid  = H5Screate_simple(1, adim, admx);
    hid_t attr = H5Acreate1(dest, name, type, aid, H5P_DEFAULT );
    H5Awrite(attr, type, buf);
    H5Sclose(aid);
    H5Aclose(attr);
}
/**
 * Creates a chuncked "series" of objects with ncol columns
 * Compression will be maximum
 *
 * tid : type
 */
static
hid_t create_array(hid_t fileid, const char *path, hid_t type, int ncol)
{
    hsize_t h_ncol    = ncol;
    hsize_t sH[2]     = { 0, h_ncol};
    hsize_t mxH[2]    = { H5S_UNLIMITED,  h_ncol};
    hsize_t chunkH[2] = {128, h_ncol};

    int rank = (ncol>1 ? 2 : 1);
    hid_t dsh = H5Screate_simple (rank, sH, mxH);
    hid_t pH  = H5Pcreate(H5P_DATASET_CREATE);
    H5Pset_layout(pH, H5D_CHUNKED);
    H5Pset_chunk(pH, rank, chunkH);
    H5Pset_deflate(pH, 4);
    hid_t tid = H5Tcopy(type);
    hid_t ret = H5Dcreate2(fileid, path,
                           tid, dsh, H5P_DEFAULT,
                           pH, H5P_DEFAULT);
    H5Sclose(dsh);
    H5Pclose(pH);
    H5Tclose(tid);

    return ret;
}

static
void append_data(hid_t data_set, int offset, int ncols, const void *buf)
{
    /*
     * Create the memory data space.
     * The one we want to add
     */
    int rank = (ncols>1 ? 2 : 1);
    hsize_t h_ncols = ncols;
    hsize_t mdims[2] = {1, h_ncols};
    hid_t m_sid = H5Screate_simple(rank, mdims, NULL);

    /*
     * Now we extend the dataset
     */
    hsize_t h_offset = offset + 1;
    hsize_t dims[2] = { h_offset, h_ncols };
    H5Dset_extent(data_set, dims);

    /*
     * ...and select the hyperslab in the file dataset
     */
    hsize_t hs_offset = offset;
    hsize_t start[2] = { hs_offset, 0 };
    hsize_t count[2] = { 1, h_ncols };
    hid_t sid = H5Dget_space(data_set); /* A copy of data space for writing */
    H5Sselect_hyperslab(sid, H5S_SELECT_SET, start, NULL, count, NULL);

    /*
     * write buffer to data space
     */
    hid_t tid = H5Dget_type(data_set);
    H5Dwrite(data_set, tid, m_sid, sid, H5P_DEFAULT, buf);
    H5Tclose(tid);
    H5Sclose(m_sid);
    H5Sclose(sid);
}

HDFDataFile::HDFDataFile(const std::string &name)
    : DataFile(name), priv (0)
{
    // TODO Auto-generated constructor stub
    priv = new HDFPrivate;
}

HDFDataFile::~HDFDataFile()
{
    if (priv)
    {
        priv->mutex.lock();
        priv->mutex.unlock();
        if ( priv->fileid>0 )
        {
            close();
        }
        delete priv;
    }
}

void HDFDataFile::open(const std::string &nam)
{
    if (!priv)
        return;

    if (priv->fileid>0)
        close();

    if (!nam.empty())
        set_file_name( nam );

    if ( PathUtils::exists(file_name()))
        unlink(file_name().c_str());

    priv->mutex.lock();
    priv->fileid =  H5Fcreate(file_name().c_str(), H5F_ACC_TRUNC, H5P_DEFAULT, H5P_DEFAULT);
    if( priv->fileid < 0 )
    {
        std::cout << "HDFDataFile: Couldn't create file " << nam << std::endl;
        priv->mutex.unlock();
        return;
    }

    /**
     *  Now we create the structure of the file
     */

    /*
     * The Date type
     */
    hid_t tDate = H5Tcreate(H5T_COMPOUND, sizeof(struct tm));
    H5Tinsert(tDate, "second", HOFFSET(tm, tm_sec), H5T_NATIVE_INT32);
    H5Tinsert(tDate, "minute", HOFFSET(tm, tm_min), H5T_NATIVE_INT32);
    H5Tinsert(tDate, "hour", HOFFSET(tm, tm_hour), H5T_NATIVE_INT32);
    H5Tinsert(tDate, "mday", HOFFSET(tm, tm_mday), H5T_NATIVE_INT32);
    H5Tinsert(tDate, "month", HOFFSET(tm, tm_mon), H5T_NATIVE_INT32);
    H5Tinsert(tDate, "year", HOFFSET(tm, tm_year), H5T_NATIVE_INT32);


    hid_t tHeader = H5Tcreate(H5T_COMPOUND, sizeof(Header));
    H5Tinsert(tHeader, "run_type", HOFFSET(Header, run_type), H5T_NATIVE_UINT32);
    H5Tinsert(tHeader, "nchips", HOFFSET(Header, nchips), H5T_NATIVE_UINT16);
    H5Tinsert(tHeader, "chip_mask", HOFFSET(Header, chip_mask), H5T_NATIVE_UINT16);
    H5Tinsert(tHeader, "firmware", HOFFSET(Header, firmware), H5T_NATIVE_UINT16);
    H5Tinsert(tHeader, "time", HOFFSET(Header, time), tDate);

    time_t tim = time(0);
    memcpy(&priv->H.time, gmtime(&tim), sizeof (struct tm)); //, &priv->H.time);


    /*
     * Header
     */
    priv->header    = H5Gcreate(priv->fileid, "/header", H5P_DEFAULT, H5P_DEFAULT, H5P_DEFAULT);
    priv->pedID     = create_array(priv->fileid, "/header/pedestal", H5T_NATIVE_FLOAT, 128*priv->H.nchips);
    priv->noiseID   = create_array(priv->fileid, "/header/noise", H5T_NATIVE_FLOAT, 128*priv->H.nchips);
    add_attribute(priv->header, "setup", tHeader, &priv->H);

    /*
     * The event group
     */
    priv->nevts = 0;
    priv->data_group = H5Gcreate(priv->fileid, "/events", H5P_DEFAULT, H5P_DEFAULT, H5P_DEFAULT);

    priv->headerID = create_array(priv->fileid, "/events/header", H5T_NATIVE_UINT16, 16*priv->H.nchips);
    priv->dataID   = create_array(priv->fileid, "/events/signal", H5T_NATIVE_UINT16,128*priv->H.nchips);
    priv->clockID  = create_array(priv->fileid, "/events/clock", H5T_NATIVE_UINT32, 1);
    priv->timeID   = create_array(priv->fileid, "/events/time", H5T_NATIVE_FLOAT, 1);
    priv->tempID   = create_array(priv->fileid, "/events/temperature", H5T_NATIVE_FLOAT, 1);


    /**
     * Create another data set with a table showing the index in the event data set where the scan point
     * with value value starts and ends
     *
     * struct { double value; int start; int end; }
     *
     */
    priv->scanID = H5Gcreate(priv->fileid, "/scan", H5P_DEFAULT, H5P_DEFAULT, H5P_DEFAULT);
    priv->nscan = 0;
    priv->scvalID = create_array(priv->fileid, "/scan/value", H5T_NATIVE_FLOAT, 1);
    priv->startID = create_array(priv->fileid, "/scan/start", H5T_NATIVE_UINT32, 1);
    priv->endID   = create_array(priv->fileid, "/scan/end",   H5T_NATIVE_UINT32, 1);

    priv->mutex.unlock();
}

// Closes the file
void HDFDataFile::close()
{
    priv->mutex.lock();

    if (priv && priv->fileid>0)
    {
        if (priv->user)
        {
            delete priv->user;
            priv->user=0;
        }

#if H5_VERSION_GE(1,8,13)
        H5Dclose(priv->headerID);
        H5Dclose(priv->dataID);
        H5Dclose(priv->tempID);
        H5Dclose(priv->timeID);
        H5Dclose(priv->clockID);
        H5Gclose(priv->data_group);


        H5Dclose(priv->scvalID);
        H5Dclose(priv->startID);
        H5Dclose(priv->endID);
        H5Gclose(priv->scanID);

        H5Dclose(priv->pedID);
        H5Dclose(priv->noiseID);
        H5Gclose(priv->header);

        if (priv->userG != H5I_BADID)
        {
            H5Dclose(priv->userID);
            H5Gclose(priv->userG);
        }
#else
        H5Fflush(priv->fileid, H5F_SCOPE_GLOBAL);
#endif
        if ( H5Fclose(priv->fileid) < 0 )
        {
            std::cout << "Could not close " << file_name() << std::endl;
        }

        priv->fileid = H5I_BADID;
    }

    priv->mutex.unlock();
}

void HDFDataFile::set_data_format(const char *fmt)
{
    priv->mutex.lock();

    if (priv && priv->fileid>0)
    {
        if ( priv->user )
            delete priv->user;

        priv->user = new UserData(priv->fileid, "userdata", fmt);
    }
    priv->mutex.unlock();
}

// Writes a user defined buffer
void HDFDataFile::write(BlockType type, size_t size, const char *buff)
{
    priv->mutex.lock();

    if (priv && priv->fileid>0 && priv->user)
    {
        /**
         * TODO: only DataBlocks
         */
        if (type == DataBlock )
            priv->user->write_data(buff);
    }

    priv->mutex.unlock();

}

// Writes the current event
void HDFDataFile::write(const AlibavaData *evt)
{

    priv->mutex.lock();

    /*
     * Copy the data in the local structure
     * TODO: Change the EventBlock so that we avoid this extra copy
     */
    struct AlibavaDataBlock data;
    const EventBlock *B = evt->get_block();
    data.clock = B->clock;
    data.temp = 0.12*evt->temp() - 39.8;

    // Decode time from TDC
    data.time = evt->time();

    if (priv->H.nchips == 1)
    {
        memcpy(&data.header, B->chip[0].header, 16*sizeof(unsigned short));
        memcpy(&data.data,   B->chip[0].signal, 128*sizeof(unsigned short));

    }
    else
    {
        for (int ichip=0; ichip<priv->H.nchips; ichip++)
        {
            memcpy(&data.header[16*ichip], B->chip[ichip].header, 16*sizeof(unsigned short));
            memcpy(&data.data[128*ichip], B->chip[ichip].signal, 128*sizeof(unsigned short));
        }
    }


    /* Clock */
    append_data(priv->clockID, priv->nevts, 1, &data.clock);

    /* Temperature */
    append_data(priv->tempID, priv->nevts, 1, &data.temp);

    /* Time */
    append_data(priv->timeID, priv->nevts, 1, &data.time);

    /* Headers */
    append_data(priv->headerID, priv->nevts, 16*priv->H.nchips, &data.header);

    /* data */
    append_data(priv->dataID, priv->nevts, 128*priv->H.nchips, &data.data);


    priv->nevts++;

    priv->mutex.unlock();

}

/** Writes the current value of pedestals and noise
 *  Also stores the run type, given in type
 */
void HDFDataFile::write(int type, double *p, double *n)
{
    if (!p || !n)
        return;

    priv->mutex.lock();

    if ( priv->H.chip_mask == 3 )
    {
        if (p)
            for (int i=0; i<256; i++)
                priv->H.pedestals[i] = p[i];

        if (n)
            for (int i=0; i<256; i++)
                priv->H.noise[i] = n[i];

    }
    else
    {
        int ioff = priv->H.chip_mask==1 ? 0 : 128;

        if (p)
            for (int i=0; i<128; i++)
                priv->H.pedestals[i] = p[ioff+i];

        if (n)
            for (int i=0; i<128; i++)
                priv->H.noise[i] = n[ioff+i];

    }

    append_data(priv->pedID,   0, 128*priv->H.nchips, priv->H.pedestals);
    append_data(priv->noiseID, 0, 128*priv->H.nchips, priv->H.noise);

    priv->mutex.unlock();

}

void HDFDataFile::new_point(int start, int end, double value)
{
    priv->mutex.lock();

    priv->P.start = start;
    priv->P.end = end;
    priv->P.value = value;

    append_data(priv->scvalID, priv->nscan, 1, &priv->P.value);
    append_data(priv->startID, priv->nscan, 1, &priv->P.start);
    append_data(priv->endID,  priv->nscan, 1, &priv->P.end);
    priv->nscan++;

    priv->mutex.unlock();
}

void HDFDataFile::scan_definition(int type, int charge, int delay, int npoints, int from, int to, int nevts)
{
    priv->mutex.lock();

    priv->H.scan.type  = type;
    priv->H.scan.npts  = npoints;
    priv->H.scan.from  = from;
    priv->H.scan.to    = to;
    priv->H.scan.nevts = nevts;
    priv->H.scan.charge = charge;
    priv->H.scan.delay = delay;

    if ( priv->fileid != H5I_BADID )
    {
        if ( !H5Aexists(priv->scanID, "scan_definition") )
        {
            hid_t tScanDef = H5Tcreate(H5T_COMPOUND, sizeof(ScanDef));
            H5Tinsert(tScanDef, "type", HOFFSET(ScanDef, type), H5T_NATIVE_UINT32);
            H5Tinsert(tScanDef, "from", HOFFSET(ScanDef, from), H5T_NATIVE_UINT32);
            H5Tinsert(tScanDef, "to", HOFFSET(ScanDef, to), H5T_NATIVE_UINT32);
            H5Tinsert(tScanDef, "npts", HOFFSET(ScanDef, npts), H5T_NATIVE_UINT32);
            H5Tinsert(tScanDef, "nevts", HOFFSET(ScanDef, nevts), H5T_NATIVE_UINT32);
            H5Tinsert(tScanDef, "charge", HOFFSET(ScanDef, charge), H5T_NATIVE_UINT32);
            H5Tinsert(tScanDef, "delay", HOFFSET(ScanDef, delay), H5T_NATIVE_UINT32);

            add_attribute(priv->scanID,"scan_definition", tScanDef, &priv->H.scan);
        }
    }
    priv->mutex.unlock();

}
void HDFDataFile::describe_setup(int run_type, int nchip, int chip_mask, int firmware)
{
    priv->mutex.lock();

    priv->H.run_type = run_type;
    priv->H.nchips = nchip;
    priv->H.chip_mask = chip_mask;
    priv->H.firmware = firmware;

    priv->mutex.unlock();
}

void HDFDataFile::flush()
{
    if (priv->fileid != H5I_BADID)
    {
        priv->mutex.lock();
        H5Fflush(priv->fileid, H5F_SCOPE_GLOBAL);
        priv->mutex.unlock();
    }
}



UserData::UserData(hid_t file_id, const char *group_name, const char *fmt)
    : fileid(file_id), data_group(H5I_BADID), data_size(0), nevts(0), gname(group_name)
{
    char cbuff[256];
    if (fileid == H5I_BADID)
        return;

    if (parse_data(fmt))
        return;

    /**
     * prepare the file
     */
    // Create the group and the data structure
    data_group = H5Gcreate(fileid, group_name, H5P_DEFAULT, H5P_DEFAULT, H5P_DEFAULT);
    if (data_group==H5I_BADID)
        return;

    ComponentList::iterator cp;
    for (cp=components.begin(); cp!=components.end(); ++cp)
    {
        sprintf(cbuff, "/%s/%s", group_name, cp->name);
        std::cout << cbuff << "<" << cp->fmt << ">[" << cp->len << "] offset: " << cp->offset << " align " << cp->align << std::endl;
        cp->obj = create_array(fileid, cbuff, cp->type, cp->len);
    }

}

UserData::~UserData()
{
    Close();
}

void UserData::Close()
{
    if (error())
        return;

    ComponentList::iterator cp;
    for (cp=components.begin(); cp!=components.end(); ++cp)
    {
        std::cout << "Closing " << cp->name << std::endl;
        H5Dclose(cp->obj);
    }
    components.clear();
    H5Gclose(data_group);
    data_group = H5I_BADID;
}

bool UserData::parse_data(const char *fmt)
{
    TokenList tokens;
    stringtok(tokens, fmt, ",");

    /*
     * Trim tokens
     */
    std::transform(tokens.begin(), tokens.end(), tokens.begin(), trim_str);

    /*
     * iterate
     */
    bool has_error = true;
    TokenList::iterator ip;
    for (ip = tokens.begin(); ip!=tokens.end(); ++ip)
    {
        FormatDef comp;
        std::string::size_type sep;
        std::string::size_type beg_num, end_num, stype;
        std::string name, type;

        has_error = true;

        sep = ip->find('/');
        if ( sep == std::string::npos )
        {
            std::cout << "Token " << *ip << " does not have type defined" << std::endl;
            break;
        }
        name = ip->substr(0, sep);
        type = ip->substr(sep+1);

        /*
         *  Now we parse the type string
         */
        /// Ordering
        Ordering ord = native;
        sep = type.find_first_of("<=>");
        if (sep==0)
        {
            switch (type[0])
            {
                case '=':
                    ord = native;
                    break;
                case '<':
                    ord = little_endian;
                    break;
                case '>':
                    ord = big_endian;
                    break;
            }
            type.erase( type.begin() );
        }
        else if (sep != std::string::npos)
        {
            std::cout << "Wrong syntax for the type: " << name << " -> " << type << std::endl;
        }

        /// Dimension
        beg_num = type.find_first_of("0123456789");
        if (beg_num != std::string::npos)
        {
            if (beg_num != 0)
            {
                std::cout << "Wrong syntax for the type: " << name << " -> " << type << std::endl;
                break;
            }
            end_num = type.find_first_not_of("0123456789");
            if (end_num == std::string::npos)
            {
                std::cout << "Wrong syntax for the type: " << name << " -> " << type
                          << ". Apparently there is no type definition."
                          << std::endl;
                break;
            }
            std::string dim = type.substr(beg_num, end_num-beg_num);
            type.erase(beg_num, end_num-beg_num);
            comp.len = atoi(dim.c_str());
        }
        else
            comp.len = 1;

        /// Type
        stype = type.find_first_of("chHiIlLfd");
        if (stype == std::string::npos)
        {
            std::cout << "Wrong syntax for the type: " << name << " -> " << type << std::endl;
            break;
        }
        else if (stype != 0 )
        {
            std::cout << "Wrong syntax for the type: " << name << " -> " << type << std::endl;
            break;
        }
        if ( type.length() != 1 )
        {
            std::cout << "Wrong syntax for type: " << type << std::endl
                      << "    there are extra characters beyond the type"
                      << std::endl;
            break;
        }

        sprintf(comp.name, "%s", name.c_str());
        comp.fmt = type[0];
        switch (type[0])
        {
            case 'c':
            case 'b':
                SETUP_TYPE(comp.type, ord,
                           H5T_NATIVE_CHAR,
                           H5T_STD_B8LE,
                           H5T_STD_B8BE);
                comp.align =
                comp.size = sizeof(char);
                break;
            case 'B':
                SETUP_TYPE(comp.type, ord,
                           H5T_NATIVE_UCHAR,
                           H5T_STD_U8LE,
                           H5T_STD_U8BE);
                comp.size = sizeof(unsigned char);
                break;
            case 'h':
                SETUP_TYPE(comp.type, ord,
                           H5T_NATIVE_INT16,
                           H5T_STD_I16LE,
                           H5T_STD_I16BE);
                comp.align = SHORT_ALIGN;
                comp.size = sizeof(int16_t);
                break;
            case 'H':
                SETUP_TYPE(comp.type, ord,
                           H5T_NATIVE_UINT16,
                           H5T_STD_U16LE,
                           H5T_STD_U16BE);
                comp.align = SHORT_ALIGN;
                comp.size = sizeof(uint16_t);
                break;
            case 'i':
                SETUP_TYPE(comp.type, ord,
                           H5T_NATIVE_INT32,
                           H5T_STD_I32LE,
                           H5T_STD_I32BE);
                comp.align = INT_ALIGN;
                comp.size = sizeof(int32_t);
                break;
            case 'I':
                SETUP_TYPE(comp.type, ord,
                           H5T_NATIVE_UINT32,
                           H5T_STD_U32LE,
                           H5T_STD_U32BE);
                comp.align = INT_ALIGN;
                comp.size = sizeof(uint32_t);
                break;
            case 'l':
                SETUP_TYPE(comp.type, ord,
                           H5T_NATIVE_INT64,
                           H5T_STD_I64LE,
                           H5T_STD_I64BE);
                comp.align = LONG_ALIGN;
                comp.size = sizeof(int64_t);
                break;
            case 'L':
                SETUP_TYPE(comp.type, ord,
                           H5T_NATIVE_UINT64,
                           H5T_STD_U64LE,
                           H5T_STD_U64BE);
                comp.align = LONG_ALIGN;
                comp.size = sizeof(uint64_t);
                break;
            case 'f':
                SETUP_TYPE(comp.type, ord,
                           H5T_NATIVE_FLOAT,
                           H5T_IEEE_F32LE,
                           H5T_IEEE_F32BE);
                comp.align = FLOAT_ALIGN;
                comp.size = sizeof(float);
                break;
            case 'd':
                SETUP_TYPE(comp.type, ord,
                           H5T_NATIVE_DOUBLE,
                           H5T_IEEE_F64LE,
                           H5T_IEEE_F64BE);
                comp.align = DOUBLE_ALIGN;
                comp.size = sizeof(double);
                break;
        }
        has_error = false;
        components.push_back(comp);
    }
    if (has_error)
        return true;


    /**
     * Compute chunk size
     */
    data_size=0;
    ComponentList::iterator cp;
    for (cp=components.begin(); cp!=components.end(); ++cp)
    {
        cp->offset = data_size;
        if (cp->align && data_size > 0)
        {
            size_t xtra = (cp->align - 1) - (data_size -1) % (cp->align);
            data_size += xtra;
            cp->offset +=xtra;
        }
        data_size += cp->len * cp->size;
    }

    return false;
}


bool UserData::write_data(const void *ptr)
{
    /**
     * Now write
     * Create a buffer of size data_size make a switch and generate a
     * value according to the type and write to the buffer.
     */
    ComponentList::iterator cp;
    for (cp=components.begin(); cp!=components.end(); ++cp)
    {
        const volatile char *ppp = (reinterpret_cast<const volatile char *>(ptr) + cp->offset);
        append_data(cp->obj, nevts, cp->len,  (const void *)ppp);
    }
    nevts++;
    return false;
}


#else
struct HDFPrivate
{
        int id;
};


HDFDataFile::HDFDataFile(const std::string &name)
    : DataFile(name), priv(0)
{
    // TODO Auto-generated constructor stub
}

HDFDataFile::~HDFDataFile()
{
    // TODO Auto-generated destructor stub
}

void HDFDataFile::open(const std::string &nam)
{

}
// Closes the file
void HDFDataFile::close()
{

}

void HDFDataFile::set_data_format(const char *fmt)
{
}

// Writes a user defined buffer
void HDFDataFile::write(BlockType type, size_t size, const char *buff)
{

}

// Writes the current event
void HDFDataFile::write(const AlibavaData *evt)
{

}

/** Writes the current value of pedestals and noise
 *  Also stores the run type, given in type
 */
void HDFDataFile::write(int type, double *p, double *n)
{

}
void HDFDataFile::scan_definition(int type, int charge, int delay, int npoints, int from, int to, int nevts)
{

}
void HDFDataFile::describe_setup(int run_type, int nchip, int chip_mask, int firmware)
{
}
void HDFDataFile::new_point(int start, int end, double value)
{

}
void HDFDataFile::flush()
{

}

#endif
